'''
Defining the disc etc.

Rewrite to use np.in1d to get the correct particles.

Examples:
    >>> from ..snapshot import Snapshot
    >>> from ..transformation import *
    >>> from ..analysis import *
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_470', physical=False)
    >>> Translation(-shrinking_sphere(s.stars, [s.boxsize/2]*3,
    ...                               np.sqrt(3)*s.boxsize)).apply(s)
    load block pos... done.
    do a shrinking sphere...
      starting values:
        center = [ 36000.  36000.  36000.] [ckpc h_0**-1]
        R      = 1.247077e+05 [ckpc h_0**-1]
    load block mass... done.
    done.
    apply Translation to "pos" of "snap_M1196_4x_470"... done.
    >>> s['vel'] -= mass_weighted_mean(s[s['r']<'1 kpc'], 'vel')
    load block vel... done.
    derive block r... done.
    >>> orientate_at(s[s['r'] < '10 kpc'].baryons, 'L', total=True)
    derive block momentum... done.
    derive block angmom... done.
    apply Rotation to "vel" of "snap_M1196_4x_470"... done.
    apply Rotation to "pos" of "snap_M1196_4x_470"... done.
    >>> sub = s[BallMask('60 kpc',sph_overlap=True)]    # 60 kpc ~ 30% R200 (R200 ~ 211 kpc)
    derive block r... done.
    load block hsml... done.
    >>> sub
    <Snap "snap_M1196_4x_470":Ball(r=60.0 [kpc]); N=198,250; z=0.000>
    >>> if abs(sub['r'].max().in_units_of('Mpc',subs=s) - 9.905) > 0.01:
    ...     print(sub['r'].max().in_units_of('Mpc',subs=s))
    >>> no_gas_max_r = sub.SubSnap(list(set(range(6))-set(gadget.families['gas'])))['r'].max()
    >>> if abs( no_gas_max_r.in_units_of('kpc',subs=s)  -  60.0 ) > 0.01:
    ...     print(no_gas_max_r)
    >>> s.to_physical_units()
    >>> sub = s[BoxMask('120 kpc',sph_overlap=False)]
    >>> sub # doctest:+ELLIPSIS
    <Snap "snap_M1196_4x_470":Box([[-60,60],[-60,60],[-60,60]] [kpc],strict); N=218,98...; z=0.000>
    >>> if np.linalg.norm( np.abs(sub['pos']).max(axis=0) - [120/2]*3 ) > 0.1:
    ...     print(np.abs(sub['pos']).max(axis=0))

    >>> discmask = DiscMask(0.85,rmax='60.0 kpc')
    >>> disc = s[discmask]
    derive block momentum... done.
    derive block angmom... done.
    derive block Ekin... done.
    derive block jcirc... done.
    derive block jzjc... done.
    derive block rcyl... done.
    >>> disc
    <Snap "snap_M1196_4x_470":Disc(jzjc<0.85,rcyl<60.0 [kpc],z<5.0 [kpc]); N=36,419; z=0.000>
    >>> gal = s[s['r']<'60 kpc']
    >>> np.array(disc.parts,float) / np.array(gal.parts)
    array([0.78431754,  0.02731191,         nan,         nan,  0.26441513,
                   nan])
    >>> bulge = gal[discmask.inverted()]
    >>> bulge
    <Snap "snap_M1196_4x_470":masked:~Disc(jzjc<0.85,rcyl<60.0 [kpc],z<5.0 [kpc]); N=159,918; z=0.000>
    >>> float(len(disc)) / len(bulge)
    0.2277354644255181

    >>> IDs = bulge['ID'][:1000:3]
    load block ID... done.
    >>> sub = gal[IDMask(IDs)]
    >>> sub
    <Snap "snap_M1196_4x_470":masked:IDMask; N=334; z=0.000>
    >>> assert np.all( sub['ID'] == IDs )
    >>> assert np.all( bulge[:1000:3]['pos'] == sub['pos'] )
    >>> antisub = gal[~IDMask(IDs)]
    >>> assert not bool( set(antisub['ID']).intersection(sub['ID']) )
    >>> assert len(antisub) + len(sub) == len(gal)
    >>> assert set(sub['ID']) == set(gal[IDMask(set(IDs))]['ID'])

    >>> expr_mask = ExprMask("abs(vel[:,2]) < '100 km/s'")
    >>> sub = gal[expr_mask]
    >>> sub
    <Snap "snap_M1196_4x_470":masked:ExprMask("abs(vel[:,2]) < '100 km/s'"); N=145,157; z=0.000>
    >>> assert np.all(sub['ID']
    ...                 == gal['ID'][np.abs(gal['vel'][:,2]) < '100 km/s'])
    >>> assert len(sub) + len(gal[~expr_mask]) == len(gal)
    >>> sub = gal[ExprMask("(abs(vel[:,2]) < '100 km/s') & (r < '30 kpc')")]
    >>> assert np.all(sub['ID'] == gal['ID'][
    ...     (np.abs(gal['vel'][:,2]) < '100 km/s') & (gal['r'] < '30 kpc')])

    Combining masks with (bitwise) logical operators:
    >>> s[expr_mask & discmask]
    <Snap "snap_M1196_4x_470":CompoundMask(ExprMask("abs(vel[:,2]) < '100 km/s'") & Disc(jzjc<0.85,rcyl<60.0 [kpc],z<5.0 [kpc])); N=36,072; z=0.000>
    >>> s[expr_mask | discmask]
    <Snap "snap_M1196_4x_470":CompoundMask(ExprMask("abs(vel[:,2]) < '100 km/s'") | Disc(jzjc<0.85,rcyl<60.0 [kpc],z<5.0 [kpc])); N=1,901,758; z=0.000>
    >>> s[expr_mask ^ discmask]
    <Snap "snap_M1196_4x_470":CompoundMask(ExprMask("abs(vel[:,2]) < '100 km/s'") ^ Disc(jzjc<0.85,rcyl<60.0 [kpc],z<5.0 [kpc])); N=1,865,686; z=0.000>
    >>> s[expr_mask & discmask & BallMask('30 kpc')]
    <Snap "snap_M1196_4x_470":CompoundMask(CompoundMask(ExprMask("abs(vel[:,2]) < '100 km/s'") & Disc(jzjc<0.85,rcyl<60.0 [kpc],z<5.0 [kpc])) & Ball(r=30.0 [kpc],strict)); N=34,177; z=0.000>
    >>> np.round( np.percentile(
    ...     s[ BallMask('100 kpc') & ~BallMask('30 kpc') ]['r'], [0,100]) )
    array([ 30.,  100.])
'''
__all__ = ['SnapMask', 'BallMask', 'BoxMask', 'DiscMask', 'IDMask', 'ExprMask']

import numpy as np
from ..units import *
from .. import gadget
#from .snapshot import SubSnap, SubSnap is now method of Snapshot
import operator


class SnapMask(object):
    '''The base class for more complicated masks.'''

    def __init__(self, inverse=False):
        self._inverse = bool(inverse)

    def is_inverse(self):
        return self._inverse

    def inverted(self):
        '''Return the inverse of this mask.'''
        raise NotImplementedError()

    def __invert__(self):
        return self.inverted()

    def __str__(self):
        s = '~' if self._inverse else ''
        s += self.__class__.__name__
        return s

    def _get_mask_for(self, s):
        '''
        The actual implementation for getting the mask for this class, not
        accounting for the _inverse flag.
        '''
        raise NotImplementedError('This is just the interface!')

    def get_mask_for(self, s):
        '''
        Actually get the mask for a given snapshot.

        Args:
            s (Snap):   The (sub-)snapshot to mask.

        Returns:
            mask (np.ndarray[bool]):    The mask to apply.
        '''
        mask = self._get_mask_for(s)
        return ~mask if self._inverse else mask

    def __and__(self, other):
        return CompoundMask(self, operator.and_, other)

    def __or__(self, other):
        return CompoundMask(self, operator.or_, other)

    def __xor__(self, other):
        return CompoundMask(self, operator.xor, other)


class BallMask(SnapMask):
    '''
    A mask for all particles within a given radius.

    Args:
        R (UnitScalar):         The maximum radius.
        center (UnitQty):       The center of the ball. Default: origin.
        sph_overlap (bool):     If True, also include gas particles, that actually
                                lie outside of the ball of radius R, but they are
                                smoothed into it.
        periodic_snap (bool):   Whether to consider the snapshot to be masked
                                periodic with its boxside or not.
    '''

    def __init__(self, R, center=None, sph_overlap=False, periodic_snap=True):
        super(BallMask, self).__init__()
        self.R = R
        self.center = center
        self.sph_overlap = sph_overlap
        self.periodic_snap = periodic_snap

    def inverted(self):
        inv = BallMask(self._R, self._center, sph_overlap=self.sph_overlap)
        inv._inverse = not self._inverse
        return inv

    @property
    def R(self):
        return self._R

    @R.setter
    def R(self, value):
        self._R = UnitScalar(value)

    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, value):
        if value is not None:
            self._center = UnitQty(value).copy()
        else:
            self._center = UnitQty([0] * 3)

    def __str__(self):
        s = '~' if self._inverse else ''
        s += 'Ball('
        if not np.all(self._center == 0):
            s += 'center=%s,' % self._center
        s += 'r=%s' % self._R
        if not self.sph_overlap:
            s += ',strict'
        return s + ')'

    def _get_mask_for(self, s):
        from ..utils import periodic_distance_to, dist
        R = self._R.in_units_of(s['r'].units, subs=s)
        center = self._center.in_units_of(s['r'].units, subs=s)

        if self.periodic_snap:
            r = periodic_distance_to(s['pos'], center, s.boxsize)
        else:
            r = dist(s['pos'], center) if not np.all(center == 0) else s['r']
        mask = r < R

        if self.sph_overlap and 'gas' in s:
            for pt in gadget.families['gas']:
                sub = s.SubSnap([pt])
                if self.periodic_snap:
                    r = periodic_distance_to(sub['pos'], center, s.boxsize)
                else:
                    r = dist(sub['pos'], center) if not np.all(center == 0) \
                        else sub['r']
                mask[sum(s.parts[:pt]):sum(s.parts[:pt + 1])] |= \
                    r - sub['hsml'] < R

        return mask.view(np.ndarray)


class BoxMask(SnapMask):
    '''
    A mask for all particles within a square box.

    The box is always a cube and aligned with the axes.

    Args:
        extent (Unit, UnitArr): The size of the box. Can either be a .
        center (UnitQty):       The center of the box, if extent is just a scalar.
                                Otherwise it will be ignored. Default: origin.
        sph_overlap (bool):     If True, also include gas particles, that actually
                                lie outside of the box, but they are smoothed
                                into it.
    '''

    def __init__(self, extent, center=None, sph_overlap=False):
        super(BoxMask, self).__init__()
        # might be needed for calculation of self.center in setting self.extent:
        self._extent = UnitArr([[-1, 1]] * 3)
        self.extent = extent
        if center is not None:
            self.center += center
        self.sph_overlap = sph_overlap

    def inverted(self):
        inv = BoxMask(self._extent, sph_overlap=self.sph_overlap)
        inv._inverse = not self._inverse
        return inv

    @property
    def extent(self):
        return self._extent

    @extent.setter
    def extent(self, value):
        extent = UnitQty(value, dtype=np.float64).copy()
        if extent.shape in [(), (3,)]:
            L = extent
            center = self.center
            extent = UnitArr(np.empty((3, 2), dtype=np.float64), extent.units)
            extent[:, 0] = center - L / 2.0
            extent[:, 1] = center + L / 2.0
        if extent.shape != (3, 2):
            raise ValueError('Extent has to ba a scalar or an array of shape ' +
                             '(3,) or (3,2), but got shape %s!' % (extent.shape,))
        self._extent = extent

    @property
    def center(self):
        return (self._extent[:, 1] + self._extent[:, 0]) / 2.0

    @center.setter
    def center(self, value):
        center_new = UnitQty(value, self._extent.units).copy()
        if center_new.shape != (3,):
            raise ValueError('Center has to have shape (3,)!')
        diff = center_new - (self._extent[:, 1] + self._extent[:, 0]) / 2.0
        self._extent[:, 0] += diff 
        self._extent[:, 1] += diff
    def __str__(self):
        s = '~' if self._inverse else ''
        s += 'Box(['
        for i in range(3):
            s += '[%.4g,%.4g]%s' % (tuple(self._extent[i]) +
                                    ('' if i == 2 else ',',))
        s += '] %s' % self._extent.units
        if not self.sph_overlap:
            s += ',strict'
        return s + ')'

    def _get_mask_for(self, s):
        ext = self._extent.in_units_of(s['pos'].units, subs=s)

        mask = (ext[0, 0] <= s['pos'][:, 0]) & (s['pos'][:, 0] <= ext[0, 1]) & \
               (ext[1, 0] <= s['pos'][:, 1]) & (s['pos'][:, 1] <= ext[1, 1]) & \
               (ext[2, 0] <= s['pos'][:, 2]) & (s['pos'][:, 2] <= ext[2, 1])

        if self.sph_overlap and 'gas' in s:
            for pt in gadget.families['gas']:
                sub = s.SubSnap([pt])
                mask[sum(s.parts[:pt]):sum(s.parts[:pt + 1])] |= \
                    (ext[0, 0] <= sub['pos'][:, 0] + sub['hsml']) & \
                    (sub['pos'][:, 0] - sub['hsml'] <= ext[0, 1]) & \
                    (ext[1, 0] <= sub['pos'][:, 1] + sub['hsml']) & \
                    (sub['pos'][:, 1] - sub['hsml'] <= ext[1, 1]) & \
                    (ext[2, 0] <= sub['pos'][:, 2] + sub['hsml']) & \
                    (sub['pos'][:, 2] - sub['hsml'] <= ext[2, 1])

        return mask.view(np.ndarray)


class DiscMask(SnapMask):
    '''
    A mask for all particles that belong to the disc.

    Here, the disc is specified by minimum jzjc (see definition of derived
    blocks!). Additionally one can require a maximum cylindrical distance form the
    center and/or a maximum modulus of the z-coordinate.

    Note:
        This definition is only sensible, if the (sub-)snapshot is already
        centered on the galaxy and orientated such that the total angular momentum
        (of the disc) points along positve z.

    Args:
        jzjc_min (float):   The minimum ratio of the z-component of the angular
                            momentum and the hypothetical circular velocity.
        rmax (UnitScalar):  An additional requirement on the cylindrical(!)
                            radius. If None, this requirement is ignored.
        zmax (UnitScalar):  An additional requirement on the z-coordinate. If
                            None, this requirement is ignored.
    '''

    def __init__(self, jzjc_min=0.85, rmax='50 kpc', zmax='5 kpc'):
        super(DiscMask, self).__init__()
        self.jzjc_min = jzjc_min
        self.rmax = rmax
        self.zmax = zmax

    def inverted(self):
        inv = DiscMask(self.jzjc_min, self._rmax, zmax=self._zmax)
        inv._inverse = not self._inverse
        return inv

    @property
    def rmax(self):
        return self._rmax

    @rmax.setter
    def rmax(self, value):
        if value is None:
            self._rmax = None
        else:
            self._rmax = UnitScalar(value)

    @property
    def zmax(self):
        return self._zmax

    @zmax.setter
    def zmax(self, value):
        if value is None:
            self._zmax = None
        else:
            self._zmax = UnitScalar(value)

    def __str__(self):
        s = '~' if self._inverse else ''
        s += 'Disc(jzjc<%.2f' % self.jzjc_min
        if self._rmax is not None:
            s += ',rcyl<%s' % self._rmax
        if self._zmax is not None:
            s += ',z<%s' % self._zmax
        return s + ')'

    def _get_mask_for(self, s):
        mask = s['jzjc'] > np.sign(np.mean(s['jzjc'])) * self.jzjc_min
        if self._rmax is not None:
            rmax = self._rmax.in_units_of(s['rcyl'].units, subs=s)
            mask &= s['rcyl'] < rmax
        if self._zmax is not None:
            zmax = self._zmax.in_units_of(s['pos'].units, subs=s)
            mask &= np.abs(s['pos'][:, 2]) < zmax
        return mask


class IDMask(SnapMask):
    '''
    Mask a snapshot to a list of given IDs.

    The given IDs do not all have to be present in the snapshot to be masked; the
    masked one, though does contain all those and only those partilces which IDs
    are given and present.

    Args:
        IDs (array-like):   IDs from previous snapshot to mask the current one.
    '''

    def __init__(self, IDs):
        super(IDMask, self).__init__()
        if isinstance(IDs, set):
            IDlist = list(IDs)
            if len(IDs) == 0:
                self._IDs = np.array(IDlist, dtype=np.uint64)
            else:
                self._IDs = np.array(IDlist)
        else:
            self._IDs = np.array(IDs, copy=True)
        return

    def inverted(self):
        inv = IDMask(self._IDs)
        inv._inverse = not self._inverse
        return inv

    @property
    def IDs(self):
        return set(self._IDs)

    def _get_mask_for(self, s):
        IDs = self._IDs
        if IDs.dtype != s['ID'].dtype:
            raise RuntimeError("You cannot mask with IDs of different type "
                               "(%s) than those of the snapshot (%s)!" % (
                                   IDs.dtype, s['ID'].dtype))
        return np.in1d(s['ID'], IDs)


class ExprMask(SnapMask):
    '''
    Mask a snapshot by a given expression.

    The expression defines a simple snapshot mask and shall be used in similar way
    as `Snap.get()`.

    Args:
        expr (str):     A python expression as it can be used in Snap.get()
                        returning a boolean array.
    '''

    def __init__(self, expr):
        super(ExprMask, self).__init__()
        self.expr = expr

    @property
    def expr(self):
        return self._expr

    @expr.setter
    def expr(self, expr):
        if not isinstance(expr, str):
            raise ValueError('`expr` must be a string')
        self._expr = expr.strip()

    def inverted(self):
        inv = ExprMask(self._expr)
        inv._inverse = not self._inverse
        return inv

    def __str__(self):
        s = '~' if self._inverse else ''
        s += 'ExprMask("%s")' % self._expr
        return s

    def _get_mask_for(self, s):
        mask = s.get(self._expr)
        return mask.view(np.ndarray)


class CompoundMask(SnapMask):
    '''
    A snapshot mask built from combining two others. Normally this class is not
    isinstantiated by the constructur, but by an operation of two others.

    Args:
        m1 (SnapMask):  The first operand.
        op (operator.and_, operand.or_, operand.xor):
                        Operation to combine the two masks.
        m2 (SnapMask):  The second operand.
    '''

    def __init__(self, m1, op, m2):
        super(CompoundMask, self).__init__()
        for m in (m1, m2):
            if not isinstance(m, SnapMask):
                raise ValueError('"%s" is not a SnapMask!' % m)
        if op not in (operator.and_, operator.or_, operator.xor):
            raise ValueError('Operator "%s" is not supported' % op)

        self._masks = [m1, m2]
        self._op = op

    @property
    def operator(self):
        return self._op

    def inverted(self):
        inv = CompoundMask(self._masks[0], self._op, self._masks[1])
        inv._inverse = not self._inverse
        return inv

    def __str__(self):
        s = '~' if self._inverse else ''
        s += 'CompoundMask(%s %s %s)' % (
            self._masks[0],
            {operator.and_: '&',
             operator.or_: '|',
             operator.xor: '^',
             }.get(self._op, str(self._op)),
            self._masks[1],
        )
        return s

    def _get_mask_for(self, s):
        m1, m2 = [m.get_mask_for(s) for m in self._masks]
        return self._op(m1, m2)




'''
Module for the pygad snapshots.

These are structures for Gadget snapshots. The reading, though, is done by the
gadget module.

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(snapshot)
    TestResults(failed=0, attempted=118)
    >>> doctest.testmod(derive_rules)
    TestResults(failed=0, attempted=4)
    >>> doctest.testmod(derived)
    TestResults(failed=0, attempted=13)
    >>> doctest.testmod(sim_arr)
    TestResults(failed=0, attempted=18)
    >>> doctest.testmod(masks)
    TestResults(failed=0, attempted=45)
'''
from .snapshot import *
from .snapshotcache import *
from .derive_rules import *     #   not clear how to import: are used often e.g. ege_from_form
from .derived import read_derived_rules
from .sim_arr import *
from .masks import *
import os

from ..environment import module_dir
read_derived_rules(['./derived.cfg',
                   os.getenv("HOME")+'/.config/pygad/derived.cfg',
                   module_dir+'config/derived.cfg'])




'''
Interface to the C implementations of binning.

Testing:
    >>> from ..environment import module_dir
    >>> from ..snapshot import Snapshot
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_320', physical=True)
    >>> from ..transformation import Translation
    >>> Translation(-UnitArr([34.7828, 35.5898, 33.6147], 'cMpc/h_0')).apply(s)
    >>> extent = UnitArr([[-0.5,0.7],[-1.0,2.0],[-2.0,2.0]], 'Mpc')
    >>> sub = s[BoxMask(extent, sph_overlap=True)]
    load block pos... done.
    apply stored Translation to block pos... done.
    load block hsml... done.
    >>> Npx = np.array([ 30,  75, 100])
    >>> map2D = SPH_to_2Dgrid(sub.gas, extent=extent[:2], qty='rho', Npx=Npx[:2])
    create a 30 x 75 SPH-grid (1200 x 3000 [kpc])...
    load block rho... done.
    load block mass... done.
    derive block dV... done.
    done with SPH grid

    Consistency check of total mass:
    >>> mask = np.ones(len(sub.gas), dtype=bool)
    >>> pos, hsml = sub.gas['pos'], sub.gas['hsml']
    >>> for k in range(3):
    ...     mask &= (extent[k,0]<pos[:,k]-hsml) & (pos[:,k]+hsml<extent[k,1])
    >>> lower = sub.gas[mask]['mass'].sum()
    >>> upper = sub.gas['mass'].sum()
    >>> del mask
    >>> if not (lower <= map2D.sum()*map2D.vol_voxel() <= upper):
    ...     print(lower, map2D.sum(), upper)

    Consistency check between 3D and 2D:
    >>> map3D  = SPH_to_3Dgrid(sub.gas, extent=extent, qty='rho', Npx=Npx)
    create a 30 x 75 x 100 SPH-grid (1200 x 3000 x 4000 [kpc])...
    done with SPH grid
    >>> map3D_proj = map3D.sum(axis=-1) * map3D.res(2,units='kpc')
    >>> tot_rel_err = (map3D_proj.sum() - map2D.sum()) / map2D.sum()
    >>> if tot_rel_err > 1e-3:
    ...     print(tot_rel_err)
    >>> px_rel_err = np.abs(map2D-map3D_proj)/map2D
    >>> if np.mean(np.abs(px_rel_err)) > 0.01:
    ...     print(np.mean(np.abs(px_rel_err)))
    >>> if np.percentile(np.abs(px_rel_err), 99) > 0.1:
    ...     print(np.percentile(np.abs(px_rel_err), 99))

    Quick consistency check of higher resolution
    >>> map2D_high = SPH_to_2Dgrid(sub.gas, extent=extent[:2], qty='rho',
    ...                            Npx=Npx[:2]*4)
    create a 120 x 300 SPH-grid (1200 x 3000 [kpc])...
    done with SPH grid
    >>> tot_rel_err = np.sum(map3D_proj - map2D) / map2D.sum()
    >>> if tot_rel_err > 0.001:
    ...     print(tot_rel_err)

    Consistency check between 3D-map and the line integral:
    >>> extent = UnitArr([[-0.1,0.1],[-0.1,0.1],[-1e2,1e2]], 'kpc')
    >>> Npx = [11,11,1100]
    >>> map3D = SPH_to_3Dgrid(sub.gas, extent=extent, qty='rho', Npx=Npx)
    create a 11 x 11 x 1100 SPH-grid (0.2 x 0.2 x 200 [kpc])...
    done with SPH grid
    >>> column = map3D[5,5,:].ravel() * map3D.res(2)
    >>> line = SPH_3D_to_line(sub.gas, qty='rho', los=[0,0],
    ...                       extent=extent[2], Npx=Npx[2])
    create a SPH-line with 1100 bins (length 200 [kpc])...
    done with SPH line
    >>> px_rel_err = np.abs(column-line)/column
    >>> if np.mean(np.abs(px_rel_err)) > 0.01:
    ...     print(np.mean(np.abs(px_rel_err)))
    >>> if np.percentile(np.abs(px_rel_err), 99) > 0.05:
    ...     print(np.percentile(np.abs(px_rel_err), 99))
'''
__all__ = ['SPH_to_3Dgrid', 'SPH_to_2Dgrid', 'SPH_3D_to_line',
           'SPH_to_2Dgrid_by_particle']

import numpy as np
from ..kernels import *
from ..units import *
from .core import *
from ..utils import *
from .. import environment
from .. import gadget
from .. import C
from numbers import Number
from ..snapshot import BoxMask

def SPH_to_3Dgrid(s, qty, extent, Npx, kernel=None, dV='dV', hsml='hsml',
                  normed=True):
    '''
    Bin some SPH quantity onto a 3D grid, fully accounting for the smoothing
    lengths.

    This method also ensures that the integral over the region of the grid is
    conserved; meaning that no particles can "fall through the grid", not even
    partially.

    Args:
        s (Snap):               The gas-only (sub-)snapshot to bin from.
        qty (UnitQty, str):     The quantity to map. It can be a UnitArr of length
                                L=len(s) and dimension 1 (i.e. shape (L,)) or a
                                string that can be passed to s.get and returns
                                such an array.
        extent (UnitQty):       The extent of the map. It can be a scalar and then
                                is taken to be the total side length of a cube
                                around the origin or a sequence of the minima and
                                maxima of the directions:
                                [[xmin,xmax],[ymin,ymax],[zmin,zmax]].
        Npx (int, sequence):    The number of pixel per side. Either an integer
                                that is taken for all three sides or a 3-tuple of
                                such, each value for one direction.
        kernel (str):           The kernel to use for smoothing. (By default use
                                the kernel defined in `gadget.cfg`.)
        dV (str, UnitQty, Unit):The volume element to use. Can be a block name, a
                                block itself or a Unit that is taken as constant
                                volume for all particles.
        hsml (str, UnitQty, Unit):
                                The smoothing lengths to use. Defined analoguous
                                to dV.
        normed (bool):          Whether to norm the kernel integral on the
                                discrete grid. This ensures the integral of the
                                binned quantity is conserved; the kernel must be
                                normed to one, of course.

    Returns:
        grid (Map):             The binned SPH quantity.
    '''
    # prepare arguments
    extent, Npx, res = grid_props(extent=extent, Npx=Npx, dim=3)
    extent = UnitQty(extent, s['pos'].units, subs=s)
    res = UnitQty(res, s['pos'].units, subs=s)
    if kernel is None:
        kernel = gadget.general['kernel']

    if res.max()/res.min() > 5:
        import sys
        print('WARNING: 3D grid has very uneven ratio of ' + \
                             'smallest to largest resolution (ratio %.2g)!' % (
                                     res.max()/res.min()), file=sys.stderr)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('create a %d x %d x %d' % tuple(Npx), end=' ')
        print('SPH-grid (%.4g x %.4g x %.4g' % tuple(extent[:,1]-extent[:,0]), end=' ')
        print('%s)...' % extent.units)

    # prepare (sub-)snapshot
    if isinstance(qty, str):
        qty = s.get(qty)
    qty_units = getattr(qty,'units',None)
    if qty_units is not None:
        qty_units = qty_units.gather()
    if qty.shape!=(len(s),):
        raise ValueError('Quantity has to have shape (N,)!')
    if len(s) == 0:
        return UnitArr(np.zeros(tuple(Npx)), qty_units), res

    if len(s.gas) not in [0,len(s)]:
        raise NotImplementedError()
    sub = s[BoxMask(extent, sph_overlap=True)]

    pos = sub['pos'].view(np.ndarray).astype(np.float64)
    if isinstance(hsml, str):
        hsml = sub[hsml].in_units_of(s['pos'].units)
    elif isinstance(hsml, (Number,Unit)):
        hsml = UnitScalar(hsml,s['pos'].units)*np.ones(len(sub), dtype=np.float64)
    else:   # should be some array
        hsml = UnitQty(hsml,s['pos'].units,subs=s)[sub._mask]
    hsml = hsml.view(np.ndarray).astype(np.float64)
    if isinstance(dV, str):
        dV = sub[dV].in_units_of(s['pos'].units**3)
    elif dV is None:
        dV = (hsml/2.0)**3
    else:
        dV = UnitArr(dV[sub._mask], s['pos'].units**3)
    dV = dV.view(np.ndarray).astype(np.float64)
    qty = qty[sub._mask].view(np.ndarray).astype(np.float64)
    if pos.base is not None:
        pos.copy()
    if hsml.base is not None:
        hsml.copy()
    if dV.base is not None:
        dV.copy()
    if qty.base is not None:
        qty.copy()

    ext = extent.view(np.ndarray).astype(np.float64).copy()
    grid = np.empty(np.prod(Npx), dtype=np.float64)
    Npx = Npx.astype(np.intp)
    if normed:
        sph_bin = C.cpygad.sph_bin_3D
    else:
        sph_bin = C.cpygad.sph_bin_3D_nonorm
    sph_bin(C.c_size_t(len(sub)),
            C.c_void_p(pos.ctypes.data),
            C.c_void_p(hsml.ctypes.data),
            C.c_void_p(dV.ctypes.data),
            C.c_void_p(qty.ctypes.data),
            C.c_void_p(ext.ctypes.data),
            C.c_void_p(Npx.ctypes.data),
            C.c_void_p(grid.ctypes.data),
            C.create_string_buffer(kernel.encode('ascii')),
            C.c_double(s.boxsize.in_units_of(s['pos'].units)),
    )
    grid = UnitArr(grid.reshape(tuple(Npx)), qty_units)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('done with SPH grid')

    return Map(grid, extent)

def SPH_to_2Dgrid(s, qty, extent, Npx, xaxis=0, yaxis=1, kernel=None, dV='dV',
                  hsml='hsml', normed=True):
    '''
    Bin some SPH quantity onto a 2D grid, fully accounting for the smoothing
    lengths.

    This method also ensures that the integral over the region of the grid is
    conserved; meaning that no particles can "fall through the grid", not even
    partially.

    Args:
        s (Snap):               The gas-only (sub-)snapshot to bin from.
        qty (UnitQty, str):     The quantity to map. It can be a UnitArr of length
                                L=len(s) and dimension 1 (i.e. shape (L,)) or a
                                string that can be passed to s.get and returns
                                such an array.
        extent (UnitQty):       The extent of the map. It can be a scalar and then
                                is taken to be the total side length of a square
                                around the origin or a sequence of the minima and
                                maxima of the directions:
                                [[xmin,xmax],[ymin,ymax]].
        Npx (int, sequence):    The number of pixel per side. Either an integer
                                that is taken for both sides or a 2-tuple of such,
                                each value for one direction.
        kernel (str):           The kernel to use for smoothing. (By default use
                                the kernel defined in `gadget.cfg`.)
        dV (str, UnitQty, Unit):The volume element to use. Can be a block name, a
                                block itself or a Unit that is taken as constant
                                volume for all particles.
        hsml (str, UnitQty, Unit):
                                The smoothing lengths to use. Defined analoguous
                                to dV.
        normed (bool):          Whether to norm the kernel integral on the
                                discrete grid. This ensures the integral of the
                                binned quantity is conserved; the kernel must be
                                normed to one, of course.

    Returns:
        grid (Map):             The binned SPH quantity.
    '''
    # prepare arguments
    zaxis = (set([0,1,2]) - set([xaxis, yaxis])).pop()
    if set([xaxis, yaxis, zaxis]) != set([0,1,2]):
        raise ValueError('Illdefined axes (x=%s, y=%s)!' % (xaxis, yaxis))
    extent, Npx, res = grid_props(extent=extent, Npx=Npx, dim=2)
    extent = UnitQty(extent, s['pos'].units, subs=s)
    res = UnitQty(res, s['pos'].units, subs=s)
    if kernel is None:
        kernel = gadget.general['kernel']

    if res.max()/res.min() > 5:
        import sys
        print('WARNING: 2D grid has very uneven ratio of ' + \
                             'smallest to largest resolution (ratio %.2g)' % (
                                     res.max()/res.min()), file=sys.stderr)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('create a %d x %d' % tuple(Npx), end=' ')
        print('SPH-grid (%.4g x %.4g' % tuple(extent[:,1]-extent[:,0]), end=' ')
        print('%s)...' % extent.units)

    # prepare (sub-)snapshot
    if isinstance(qty, str):
        qty = s.get(qty)
    qty_units = getattr(qty,'units',None)
    if qty_units is None:
        qty_units = s['pos'].units**-2
    else:
        qty_units = (qty_units * s['pos'].units).gather()   # integrated!
    if qty.shape!=(len(s),):
        raise ValueError('Quantity has to have shape (N,)!')
    if len(s) == 0:
        return UnitArr(np.zeros(tuple(Npx)), qty_units), res

    if len(s.gas) not in [0,len(s)]:
        raise NotImplementedError()
    ext3D = UnitArr(np.empty((3,2)), extent.units)
    ext3D[xaxis] = extent[0]
    ext3D[yaxis] = extent[1]
    ext3D[zaxis] = [-np.inf, +np.inf]
    sub = s[BoxMask(ext3D, sph_overlap=True)]

    # TODO: why always need a copy? C vs Fortran alignment...!?
    pos = sub['pos'].view(np.ndarray)[:,(xaxis,yaxis)].astype(np.float64).copy()
    if isinstance(hsml, str):
        hsml = sub[hsml].in_units_of(s['pos'].units)
    elif isinstance(hsml, (Number,Unit)):
        hsml = UnitScalar(hsml,s['pos'].units)*np.ones(len(sub), dtype=np.float64)
    else:   # should be some array
        hsml = UnitQty(hsml,s['pos'].units,subs=s)[sub._mask]
    hsml = hsml.view(np.ndarray).astype(np.float64)
    if isinstance(dV, str):
        dV = sub[dV].in_units_of(s['pos'].units**3)
    elif dV is None:
        dV = (hsml/2.0)**3
    else:
        dV = UnitArr(dV[sub._mask], s['pos'].units**3)
    dV = dV.view(np.ndarray).astype(np.float64)
    qty = qty[sub._mask].view(np.ndarray).astype(np.float64)
    if hsml.base is not None:
        hsml.copy()
    if dV.base is not None:
        dV.copy()
    if qty.base is not None:
        qty.copy()

    ext = extent.view(np.ndarray).astype(np.float64).copy()
    grid = np.empty(np.prod(Npx), dtype=np.float64)
    Npx = Npx.astype(np.intp)
    if normed:
        sph_bin = C.cpygad.sph_3D_bin_2D
    else:
        sph_bin = C.cpygad.sph_3D_bin_2D_nonorm
    sph_bin(C.c_size_t(len(sub)),
            C.c_void_p(pos.ctypes.data),
            C.c_void_p(hsml.ctypes.data),
            C.c_void_p(dV.ctypes.data),
            C.c_void_p(qty.ctypes.data),
            C.c_void_p(ext.ctypes.data),
            C.c_void_p(Npx.ctypes.data),
            C.c_void_p(grid.ctypes.data),
            C.create_string_buffer(kernel.encode('ascii')),
            C.c_double(s.boxsize.in_units_of(s['pos'].units)),
    )
    grid = UnitArr(grid.reshape(tuple(Npx)), qty_units)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('done with SPH grid')

    return Map(grid, extent)

def SPH_3D_to_line(s, qty, los, extent, Npx, xaxis=0, yaxis=1, kernel=None,
                   dV='dV', hsml='hsml'):
    '''
    Bin some (integrated) quantity along a line along a coordinate axis.

    Each bin of the (infinitesmal thin) line will contain the integral of the
    given quantity along this bin. (For binnning the density, for instance, that
    will be the column density of the given bin. Summing over all bins, gives the
    total column density along the line.)

    Args:
        s (Snap):               The gas-only (sub-)snapshot to bin from.
        qty (UnitQty, str):     The quantity to map. It can be a UnitArr of length
                                L=len(s) and dimension 1 (i.e. shape (L,)) or a
                                string that can be passed to s.get and returns
                                such an array.
        los (UnitQty):          TODO
        extent (UnitQty):       The extent of the map. It can be a scalar and then
                                is taken to be the total side length of a square
                                around the origin or a sequence of the minima and
                                maxima of the directions:
                                [[xmin,xmax],[ymin,ymax]].
        Npx (int, sequence):    The number of pixel per side. Either an integer
                                that is taken for both sides or a 2-tuple of such,
                                each value for one direction.
        kernel (str):           The kernel to use for smoothing. (By default use
                                the kernel defined in `gadget.cfg`.)
        dV (str, UnitQty, Unit):The volume element to use. Can be a block name, a
                                block itself or a Unit that is taken as constant
                                volume for all particles.
        hsml (str, UnitQty, Unit):
                                The smoothing lengths to use. Defined analoguous
                                to dV.

    Returns:
        line (Map):             The binned SPH quantity.
    '''
    # prepare arguments
    zaxis = (set([0,1,2]) - set([xaxis, yaxis])).pop()
    if set([xaxis, yaxis, zaxis]) != set([0,1,2]):
        raise ValueError('Illdefined axes (x=%s, y=%s)!' % (xaxis, yaxis))
    extent = UnitQty(extent, s['pos'].units, subs=s).reshape([1,2])
    extent, Npx, res = grid_props(extent=extent, Npx=Npx, dim=1)
    extent = UnitQty(extent, s['pos'].units, subs=s)
    Npx = int(Npx)
    res = UnitScalar(res.reshape([]), s['pos'].units, subs=s)
    los = UnitQty(los, s['pos'].units, dtype=np.float64, subs=s)
    if los.shape != (2,):
        raise ValueError("`los` must have shape (2,)!")
    if kernel is None:
        kernel = gadget.general['kernel']

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('create a SPH-line with %d bins' % Npx, end=' ')
        print('(length %.4g %s)...' % (extent[0,1]-extent[0,0], extent.units))

    # prepare (sub-)snapshot
    if isinstance(qty, str):
        qty = s.get(qty)
    qty_units = getattr(qty,'units',None)
    if qty_units is None:
        qty_units = s['pos'].units
    else:
        qty_units = (qty_units * s['pos'].units).gather()   # integrated!
    if qty.shape!=(len(s),):
        raise ValueError('Quantity has to have shape (N,)!')
    if len(s) == 0:
        return UnitArr(np.zeros(Npx), qty_units), res

    if len(s.gas) != len(s):
        raise NotImplementedError()
    sub = s.gas[ periodic_distance_to(s.gas['pos'][:,(xaxis,yaxis)],
                                      los, s.boxsize) < s.gas['hsml'] ]

    # TODO: why always need a copy? C vs Fortran alignment...!?
    pos = sub['pos'].view(np.ndarray)[:,(xaxis,yaxis,zaxis)].astype(np.float64).copy()
    if isinstance(hsml, str):
        hsml = sub[hsml].in_units_of(s['pos'].units)
    elif isinstance(hsml, (Number,Unit)):
        hsml = UnitScalar(hsml,s['pos'].units)*np.ones(len(sub), dtype=np.float64)
    else:   # should be some array
        hsml = UnitQty(hsml,s['pos'].units,subs=s)[sub._mask]
    hsml = hsml.view(np.ndarray).astype(np.float64)
    if isinstance(dV, str):
        dV = sub[dV].in_units_of(s['pos'].units**3)
    elif dV is None:
        dV = (hsml/2.0)**3
    else:
        dV = UnitArr(dV[sub._mask], s['pos'].units**3)
    dV = dV.view(np.ndarray).astype(np.float64)
    qty = qty[sub._mask].view(np.ndarray).astype(np.float64)
    if hsml.base is not None:
        hsml.copy()
    if dV.base is not None:
        dV.copy()
    if qty.base is not None:
        qty.copy()

    ext = extent.view(np.ndarray).reshape((2,)).astype(np.float64).copy()
    line = np.empty(Npx, dtype=np.float64)
    C.cpygad.bin_sph_along_line(C.c_size_t(len(sub)),
                                C.c_void_p(pos.ctypes.data),
                                C.c_void_p(hsml.ctypes.data),
                                C.c_void_p(dV.ctypes.data),
                                C.c_void_p(qty.ctypes.data),
                                C.c_void_p(los.ctypes.data),
                                C.c_void_p(ext.ctypes.data),
                                C.c_size_t(Npx),
                                C.c_void_p(line.ctypes.data),
                                C.create_string_buffer(kernel.encode('ascii')),
                                C.c_double(s.boxsize.in_units_of(s['pos'].units)))
    line = UnitArr(line, qty_units)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('done with SPH line')

    return Map(line, extent, Npx=Npx)

def SPH_to_2Dgrid_by_particle(s, qty, extent, Npx, reduction, xaxis=0, yaxis=1,
                              kernel=None, av=None, dV='dV', hsml='hsml'):
    '''
    Reduce a specified quantity of all particles along the third axis within each
    2D pixel of a (projected) grid on a kernel-weighted particle basis.

    TODO

    Args:
        s (Snap):               The gas-only (sub-)snapshot to bin from.
        qty (UnitQty, str):     The quantity to map. It can be a UnitArr of length
                                L=len(s) and dimension 1 (i.e. shape (L,)) or a
                                string that can be passed to s.get and returns
                                such an array.
        extent (UnitQty):       The extent of the map. It can be a scalar and then
                                is taken to be the total side length of a square
                                around the origin or a sequence of the minima and
                                maxima of the directions:
                                [[xmin,xmax],[ymin,ymax]].
        reduction (str):        The reduction method. Available are:
                                * 'mean':   Take the mean of the partivcle
                                            property `qty`.
                                * 'stddev': Take the properties standard deviation
                                            along the line of sight.
        Npx (int, sequence):    The number of pixel per side. Either an integer
                                that is taken for both sides or a 2-tuple of such,
                                each value for one direction.
        kernel (str):           The kernel to use for smoothing. (By default use
                                the kernel defined in `gadget.cfg`.)
        av (UnitQty, str):      Take this quantity to average / weight the
                                peroperty `qty` with.
        dV (str, UnitQty, Unit):The volume element to use. Can be a block name, a
                                block itself or a Unit that is taken as constant
                                volume for all particles.
        hsml (str, UnitQty, Unit):
                                The smoothing lengths to use. Defined analoguous
                                to dV.

    Returns:
        grid (Map):             The binned SPH quantity.
    '''
    # prepare arguments
    zaxis = (set([0,1,2]) - set([xaxis, yaxis])).pop()
    if set([xaxis, yaxis, zaxis]) != set([0,1,2]):
        raise ValueError('Illdefined axes (x=%s, y=%s)!' % (xaxis, yaxis))
    extent, Npx, res = grid_props(extent=extent, Npx=Npx, dim=2)
    extent = UnitQty(extent, s['pos'].units, subs=s)
    res = UnitQty(res, s['pos'].units, subs=s)
    if kernel is None:
        kernel = gadget.general['kernel']

    if res.max()/res.min() > 5:
        import sys
        print('WARNING: 2D grid has very uneven ratio of ' + \
                             'smallest to largest resolution (ratio %.2g)' % (
                                     res.max()/res.min()), file=sys.stderr)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('create a particle %d x %d' % tuple(Npx), end=' ')
        print('SPH-grid (%.4g x %.4g' % tuple(extent[:,1]-extent[:,0]), end=' ')
        print('%s) (reduction="%s")...' % (extent.units, reduction))

    # prepare (sub-)snapshot
    if isinstance(qty, str):
        qty = s.get(qty)
    qty_units = getattr(qty,'units',None)
    if qty.shape!=(len(s),):
        raise ValueError('Quantity has to have shape (N,)!')
    if isinstance(av, str):
        av = s.get(av)
    if av is not None and av.shape!=(len(s),):
        raise ValueError('Weights quantity (`av`) has to have shape (N,)!')
    if len(s) == 0:
        return UnitArr(np.zeros(tuple(Npx)), qty_units), res

    if len(s.gas) not in [0,len(s)]:
        raise NotImplementedError()
    ext3D = UnitArr(np.empty((3,2)), extent.units)
    ext3D[xaxis] = extent[0]
    ext3D[yaxis] = extent[1]
    ext3D[zaxis] = [-np.inf, +np.inf]
    sub = s[BoxMask(ext3D, sph_overlap=True)]

    # TODO: why always need a copy? C vs Fortran alignment...!?
    pos = sub['pos'].view(np.ndarray)[:,(xaxis,yaxis)].astype(np.float64).copy()
    if isinstance(hsml, str):
        hsml = sub[hsml].in_units_of(s['pos'].units)
    elif isinstance(hsml, (Number,Unit)):
        hsml = UnitScalar(hsml,s['pos'].units)*np.ones(len(sub), dtype=np.float64)
    else:   # should be some array
        hsml = UnitQty(hsml,s['pos'].units,subs=s)[sub._mask]
    hsml = hsml.view(np.ndarray).astype(np.float64)
    if isinstance(dV, str):
        dV = sub[dV].in_units_of(s['pos'].units**3)
    elif dV is None:
        dV = (hsml/2.0)**3
    else:
        dV = UnitArr(dV[sub._mask], s['pos'].units**3)
    dV = dV.view(np.ndarray).astype(np.float64)
    qty = qty[sub._mask].view(np.ndarray).astype(np.float64)
    av = None if av is None else av[sub._mask].view(np.ndarray).astype(np.float64)
    if hsml.base is not None:
        hsml.copy()
    if dV.base is not None:
        dV.copy()
    if qty.base is not None:
        qty.copy()
    if av is not None:
        if av.base is not None:
            av.copy()
    else:
        av = np.ones(len(sub), dtype=np.float64)

    ext = extent.view(np.ndarray).astype(np.float64).copy()
    grid = np.empty(np.prod(Npx), dtype=np.float64)
    Npx = Npx.astype(np.intp)

    if reduction == 'mean':
        bin_sph_reduction = C.cpygad.bin_sph_proj_mean
    elif reduction == 'median':
        bin_sph_reduction = C.cpygad.bin_sph_proj_median
    elif reduction == 'stddev':
        bin_sph_reduction = C.cpygad.bin_sph_proj_stddev
    else:
        raise ValueError("Unknown reduction method '%s'!" % reduction)

    bin_sph_reduction(C.c_size_t(len(sub)),
                      C.c_void_p(pos.ctypes.data),
                      C.c_void_p(hsml.ctypes.data),
                      C.c_void_p(dV.ctypes.data),
                      C.c_void_p(qty.ctypes.data),
                      C.c_void_p(av.ctypes.data),
                      C.c_void_p(ext.ctypes.data),
                      C.c_void_p(Npx.ctypes.data),
                      C.c_void_p(grid.ctypes.data),
                      C.create_string_buffer(kernel.encode('ascii')),
                      C.c_double(s.boxsize.in_units_of(s['pos'].units)),
    )
    grid = UnitArr(grid.reshape(tuple(Npx)), qty_units)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('done with particle SPH grid')

    return Map(grid, extent)




'''
General image manipulation routines.

Examples:
    >>> for e in grid_props('100 kpc', Npx=100, dim=2): print(e)
    [[-50.  50.]
     [-50.  50.]] [kpc]
    [100 100]
    [1. 1.] [kpc]
    >>> for e in grid_props(UnitArr([[-5,5],[-10,3]],'kpc'), Npx=20): print(e)
    [[ -5.   5.]
     [-10.   3.]] [kpc]
    [20 20]
    [0.5  0.65] [kpc]
    >>> for e in grid_props(UnitArr([[-5,5],[-10,3]],'kpc'), Npx=[20,26]): print(e)
    [[ -5.   5.]
     [-10.   3.]] [kpc]
    [20 26]
    [0.5 0.5] [kpc]
    >>> for e in grid_props(UnitArr([5,10],'kpc'), Npx=100): print(e)
    [[-2.5  2.5]
     [-5.   5. ]] [kpc]
    [100 100]
    [0.05 0.1 ] [kpc]
    >>> for e in grid_props('1 Mpc', Npx=100, dim=3): print(e)
    [[-0.5  0.5]
     [-0.5  0.5]
     [-0.5  0.5]] [Mpc]
    [100 100 100]
    [0.01 0.01 0.01] [Mpc]
    >>> for e in grid_props([1,2], Npx=[100,50]): print(e)
    [[-0.5  0.5]
     [-1.   1. ]]
    [100  50]
    [0.01 0.04]
    >>> grid_props('1 Mpc', Npx=100)
    Traceback (most recent call last):
    ...
    ValueError: Number of dimensions not given and cannot be inferred!
    >>> grid_props([1,2], Npx=[100,50,80])
    Traceback (most recent call last):
    ...
    ValueError: Dimension mismatch of the parameters!

    >>> from ..environment import module_dir
    >>> from ..snapshot import *
    >>> from ..analysis import *
    >>> from ..transformation import *
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_470')
    >>> Translation(UnitArr([-48087.1,-49337.1,-46084.3],'kpc')).apply(s)
    >>> s['vel'] -= mass_weighted_mean(s[s['r']<'1 kpc'], 'vel')
    load block vel... done.
    load block pos... done.
    apply stored Translation to block pos... done.
    derive block r... done.
    load block mass... done.
    >>> orientate_at(s[s['r'] < '10 kpc'].baryons, 'L', total=True)
    derive block momentum... done.
    derive block angmom... done.
    apply Rotation to "pos" of "snap_M1196_4x_470"... done.
    apply Rotation to "vel" of "snap_M1196_4x_470"... done.
    >>> assert np.all(gridbin1d(s['pos'][:,0], qty=s['mass'], extent= [-100,100] )
    ...             == gridbin( s['pos'][:,0].reshape((len(s),1)),
    ...                                        qty=s['mass'], extent=[[-100,100]]))
    >>> sub = s[BoxMask('100 kpc',sph_overlap=False)]
    >>> assert np.all(gridbin2d(sub['pos'][:,0],sub['pos'][:,1])
    ...             == gridbin(sub['pos'][:,(0,1)]))
    >>> m1 = gridbin2d(sub['pos'][:,0], sub['pos'][:,1], sub['mass'], bins=100)
    >>> m2 = gridbin(sub['pos'][:,(0,1)], sub['mass'], bins=100)
    >>> assert np.all(m1==m2)
    >>> mcube = gridbin(sub['pos'], sub['mass'], bins=100)
    >>> assert np.all(m2 == mcube.sum(axis=2))
    >>> m1  # doctest: +ELLIPSIS
    <Map at 0x...; units="1e+10 Msol h_0**-1", Npx=(100, 100)>
    >>> m1.extent
    UnitArr([[-35.99917603,  35.99324417],
             [-35.99769974,  35.9996376 ]])
    >>> m1.Npx
    (100, 100)
    >>> m1.res()
    UnitArr([0.7199242 , 0.71997337])
    >>> m1.vol_voxel()
    UnitArr(0.518326256290704)
    >>> m1.vol_tot()
    UnitArr(5.183263e+03)
    >>> m1 = scale01(m1, np.percentile(m1, [5,95]))
    >>> assert m1.min() >= 0 and m1.max() <= 1
'''
__all__ = ['gridbin2d', 'gridbin1d', 'gridbin', 'grid_props', 'Map', 'scale01', 'smooth']

import numpy as np
from ..units import *
from scipy.stats import binned_statistic_dd
from scipy.ndimage.filters import convolve


def gridbin2d(x, y, qty=None, bins=50, extent=None, normed=False, stats=None,
              nanval=None):
    '''
    Bin data on a 2-dim. grid.

    This calls gridbin with x and y combined to pnts. See gridbin for more
    information!
    '''
    return gridbin(np.array([x, y]).T, qty=qty, bins=bins, extent=extent,
                   normed=normed, stats=stats, nanval=nanval)


def gridbin1d(x, qty=None, bins=50, extent=None, normed=False, stats=None,
              nanval=None):
    '''
    Bin data 1-dimensional.

    This calls gridbin. See gridbin for more information!
    '''
    extent = np.asarray(extent)
    if extent.shape == (2,):
        extent = extent.reshape((1, 2))
    return gridbin(np.array(x).reshape((len(x), 1)), qty=qty, bins=bins, extent=extent,
                   normed=normed, stats=stats, nanval=nanval)


def gridbin(pnts, qty=None, bins=50, extent=None, normed=False, stats=None,
            nanval=None):
    '''
    Bin data on a grid.

    It can used to speed-up scatter plots, for instance.

    It basically is a wrapper for binned_statistic_dd for a little more
    convenience and for adding units.

    Args:
        pnts (array-like):  An (N,D)-array of the points to bin.
        qty (UnitArr):      Values for the points to bin. If None, points are just
                            counted per bin.
        bins (int, array-like):
                            Number of bins per dimension.
        extent (secquence): The range for the grid. A squence of pairs of the
                            minimum and maximum for each dimension.
                            If both this and pnts are UnitArr's, it is taken care
                            of the units.
        normed (True):      Whether to norm the gridded quantity to one.
        stats (str, function):
                            A function to apply at the end. The predefined ones
                            are: 'count', 'sum', 'mean', 'median'. For more
                            information see binned_statistic_dd.
                            Default: 'count' if `qty` is None else 'sum'
        nanval (value):     All points where the grid is NaN are set to this
                            value.

    Returns:
        gridded (Map):      The (N,...,N)-array of the binned data.
    '''
    known_stats = ['count', 'sum', 'mean', 'median']
    if isinstance(stats, str) and stats not in known_stats:
        raise ValueError('Unknown statistic. Choose from: %s' % known_stats)
    pnts = np.asanyarray(pnts)
    if len(pnts.shape) != 2:
        raise ValueError('The points array has to have shape (N,D)!')

    if qty is None:
        if stats is None or stats == 'count':
            stats = 'count'
            qty = np.ones(len(pnts), int)
    else:
        if stats is None:
            stats = 'sum'
        qty = np.asanyarray(qty)

    if isinstance(extent, UnitArr) and isinstance(pnts, UnitArr):
        extent = extent.in_units_of(pnts.units)

    gridded, edges, binnum = binned_statistic_dd(pnts, qty, range=extent,
                                                 statistic=stats, bins=bins)
    if extent is None:
        extent = np.array([[e[0], e[-1]] for e in edges])

    gridded = UnitArr(gridded)
    # if the values to bin have units, the result should as well
    if isinstance(qty, UnitArr) and not stats == 'count':
        if isinstance(stats, str):
            gridded.units = qty.units
        else:
            if stats in UnitArr._ufunc_registry:
                gridded.units = UnitArr._ufunc_registry[ufunc](qty)
            else:
                import warnings
                warnings.warn('Operation \'%s\' on units is ' % stats.__name__ + \
                              '*not* defined! Return normal numpy array.')
                gridded = gridded.view(np.ndarray)

    if normed:
        gridded /= gridded.sum()

    if nanval is not None:
        gridded[np.isnan(gridded)] = nanval

    return Map(gridded, extent=extent)


def grid_props(extent, Npx=256, dim=None):
    '''
    Calculate the grid properties from given values.

    Args:
        extent (UnitQty):   This can either be a scalar, it then defines the
                            (total) with of the grid, a list of widths for each
                            dimension, or a full extent of a sequence of maximim
                            and minimum for all coordinates:
                            [[x1min,x1max],[x2min,x2max],...].
        Npx (int, sequence):The number of pixels per side. It can either be a
                            single value that is taken for all sides or a tuple
                            with values for each direction.
        dim (int):          The number of dimensions of the grid. If possible, it
                            will be inferred from the arguments, if it is None.

    Returns:
        extent (UnitQty):   The extent as a (dim,2)-array:
                            [[x1min,x1max],[x2min,x2max],...]
        Npx (sequence):     The number of pixels per side: [<in x1>,<in x2>,...].
        res (UnitQty):      The resolution in the different directions. It might
                            differ from the one passed (if passed), since the
                            numbers of pixels have to be intergers.
    '''
    extent = UnitQty(extent, dtype=float)
    Npx = np.array(Npx, dtype=int) if Npx is not None else None
    if dim is None:
        if extent is not None and extent.shape != ():
            dim = len(extent)
        elif Npx is not None and Npx.shape != ():
            dim = len(Npx)
        else:
            raise ValueError('Number of dimensions not given and cannot be inferred!')
    for v in (extent, Npx):
        if v is not None and v.ndim != 0:
            if len(v) != dim:
                raise ValueError('Dimension mismatch of the parameters!')

    if extent.shape == ():
        widths = UnitQty([extent] * dim, getattr(extent, 'units', None))
        w2 = extent / 2.
        extent = UnitArr([[-w2, w2]] * dim, getattr(w2, 'units', None))
    elif extent.ndim == 1:
        widths = extent
        w2s = extent / 2.
        extent = UnitArr([[-w2, w2] for w2 in w2s], getattr(w2s, 'units', None))
    else:
        extent = extent.reshape((dim, 2,))
        widths = UnitArr(extent[:, 1] - extent[:, 0])
    # now extent is in final form and widths is defined

    if Npx.shape == ():
        Npx = np.array([Npx] * dim)
    # now also Npx is in final form (of shape (dim,))

    res = widths / Npx

    return extent, Npx, res


class Map(UnitArr):
    """
    An object of a grid with extent and resolution. It inherits from `UnitArr`.

    Args:
        grid (UnitQty):     The grid itself. Does not need to have the correct
                            shape yet.
        extent (UnitQty):   This can either be a scalar, it then defines the
                            (total) with of the grid, a list of widths for each
                            dimension, or a full extent of a sequence of maximim
                            and minimum for all coordinates:
                            [[x1min,x1max],[x2min,x2max],...].
        Npx (int, array-like):
                            If not None, the passed `grid` will be reshaped
                            accordingly. `extent` and `Npx` will be passed to
                            `grid_props`.
        keyword arguments:  Passed on to the array factory function of numpy
                            (np.array). By default this includes `copy=False`.
    """

    def __new__(cls, grid, extent, Npx=None, **kwargs):
        if 'copy' not in kwargs:
            kwargs['copy'] = False
        m = UnitArr(grid, **kwargs).view(cls)
        m._extent, Npx, res = grid_props(extent=extent,
                                         Npx=m.shape if Npx is None else Npx)

        Npx = tuple(Npx)
        if m.shape != Npx:
            m = m.reshape(Npx)

        return m

    def __array_finalize__(self, obj):
        UnitArr.__array_finalize__(self, obj)
        self._extent = getattr(obj, '_extent', [[0, 0]] * len(self.shape))

    def __array_wrap__(self, array, context=None):
        return UnitArr.__array_wrap__(self, array, context)

    def __getitem__(self, key):
        from warnings import simplefilter
        # ignore all future warnings
        simplefilter(action='ignore', category=FutureWarning)
        item = super(Map, self).__getitem__(key)
        # the warning: FutureWarning: Using a non-tuple sequence for multidimensional indexing is deprecated; use `arr[tuple(seq)]` instead of `arr[seq]`. In the future this will be interpreted as an array index, `arr[np.array(seq)]`, which will result either in an error or a different result.
        # is handles by pagad in the following code
        if item.ndim != self.ndim:
            if (isinstance(key, np.ndarray) and key.dtype == bool) or \
                    (not isinstance(key, slice) and
                     all(isinstance(k, bool) for k in key)):
                # TODO: handle: boolean masking, that turned out to mask out a dim'
                pass
            else:
                # TODO: how to handle np.newaxis/None?
                if not isinstance(key, (slice, tuple, list, np.ndarray)):
                    raise NotImplementedError("key/index is not of type 'tuple',"
                                              " but '%s'" % type(key))
                take = [isinstance(s, (slice, tuple, list, np.ndarray)) for s in key] + \
                       [True] * (self.ndim - len(key))
                item._extent = self._extent[take]
        return item

    def __repr__(self):
        r = self.view(UnitArr).__repr__(val='')
        r = r.replace('UnitArr(,', '<Map at 0x%x;' % id(self))
        r = r[:-1] + ', Npx=%s>' % (self.Npx,)
        # V = self.vol_tot()
        # r = r[:-1] + ', V=%g %s>' % (V, V.units)
        return r

    def __str__(self):
        s = self.view(UnitArr).__str__()
        s += ' (Npx=%s)' % (self.Npx,)
        return s

    @property
    def grid(self):
        """Just the underlying `UnitArr`."""
        return self.view(UnitArr)

    @property
    def Npx(self):
        """The side lengths in pixels in all dimensions."""
        return self.shape

    @property
    def extent(self):
        """The grid's extent in all dimensions."""
        return self._extent

    def res(self, axis=None, units=None, subs=None):
        """Resolution: pixels size in all the dimensions."""
        extent, Npx, res = grid_props(extent=self.extent, Npx=self.Npx)
        if units:
            res.convert_to(units, subs=subs)
        if axis is None:
            return res
        else:
            return UnitArr(res[axis], res.units)

    def vol_voxel(self):
        """The voxel volume."""
        return np.prod(self.res())

    def vol_tot(self):
        """The total grid volume."""
        return np.prod(self.extent.ptp(axis=1))

    def __copy__(self, *a):
        if a:
            duplicate = UnitArr.__copy__(self, *a).view(Map)
        else:
            duplicate = UnitArr.__copy__(self).view(Map)
        duplicate._extent = self._extent
        return duplicate

    def __deepcopy__(self, *a):
        duplicate = UnitArr.__deepcopy__(self).view(Map)
        duplicate._extent = self._extent
        return duplicate


def scale01(arr, lims):
    '''
    Scale the values in an array to the interval [0,1].

    Linear map onto [0,1] with the given limits. Values beyond the limits are
    mapped to the borders.

    Args:
        arr (array-like):   The array to scale.
        lims (sequence):    The minimum and maximum value to scale to.

    Returns:
        arr (array-like):   The scaled array.
    '''
    arr = (arr - float(lims[0])) / (lims[1] - lims[0])
    arr[arr > 1.0] = 1.0
    arr[arr < 0.0] = 0.0
    return arr


def smooth(grid, sml, kernel, bndrymode='constant'):
    '''
    Smooth a gridded quantity (e.g. an image).

    Args:
        grid (array-like):      The grid to smooth. Can be a UnitArr and stays
                                one.
        sml (float):            The number of pixels to smooth over (a square
                                array is used for convolution).
        kernel (vector-function):
                                The kernel to use for smoothing. It is a function
                                of one argument (the radius in px/sml) and does
                                not have to be normed (is gets normed in here).
                                It has to be a vector function, i.e. be able to
                                operate on entire arrays.
        bndrymode (str):        How to handle the boundaries. See e.g.
                                scipy.ndimage.filters.convolve for more
                                information.

    Returns:
        smooth (array-like):    The smoothed grid. It preserves units.
    '''
    if sml < 0:
        raise ValueError('The smoothing length has to be positive or zero!')
    if sml == 0:
        return grid

    sml = float(sml)

    from ..kernels import kernels, vector_kernels

    pxs = int(2 * np.ceil(sml) + 1)
    x = np.linspace(float(-(pxs - 1) / 2.), float((pxs - 1) / 2.), pxs) / sml
    D = len(grid.shape)
    x = np.meshgrid(*(x,) * D)
    x = np.array(x)
    dists = np.sqrt(np.sum(x ** 2, axis=0))
    conv_grid = kernel(dists.ravel())
    conv_grid = conv_grid.reshape((pxs, pxs))
    conv_grid /= np.sum(conv_grid)
    smooth = convolve(grid, conv_grid, mode=bndrymode)

    if isinstance(grid, UnitArr):
        smooth = UnitArr(smooth, grid.units)
    return smooth




'''
A collection of some general (low-level) functions.

Doctests are in the functions themselves.
'''
__all__ = ['DevNull', 'static_vars', 'nice_big_num_str', 'float_to_nice_latex',
           'perm_inv', 'periodic_distance_to', 'sane_slice', 'is_consecutive',
           'rand_dir', 'ProgressBar', 'time_from_color_str', 'time_to_nice_str',
           'sec_to_nice_str', 'weighted_percentile']

import numpy as np
import re
import scipy.spatial.distance
import sys
from .term import *
import time
import numbers


class DevNull():
    '''A always open write-only null file object.'''

    def __init__(self, *args, **kwargs):
        self.closed = False
        self.mode = "w"
        self.name = "<null>"
        self.encoding = None
        self.errors = None
        self.newlines = None
        self.softspace = 0

    def close(self):
        pass

    def flush(self):
        pass

    def __next__(self):
        raise IOError("Invalid operation")

    def read(size=0):
        raise IOError("Invalid operation")

    def readline(self):
        raise IOError("Invalid operation")

    def readlines(self):
        raise IOError("Invalid operation")

    def xreadlines(self):
        raise IOError("Invalid operation")

    def seek(self):
        raise IOError("Invalid operation")

    def tell(self):
        return 0

    def truncate(self):
        pass

    def write(self, *args, **kwargs):
        pass

    def writelines(self, *args, **kwargs):
        pass


def static_vars(**kwargs):
    '''
    Decorate a function with static variables.

    Example:
        >>> @static_vars(counter=0)
        ... def foo():
        ...     foo.counter += 1
        ...     print('foo got called the %d. time' % foo.counter)
        >>> foo()
        foo got called the 1. time
        >>> foo()
        foo got called the 2. time
        >>> foo()
        foo got called the 3. time
    '''

    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func

    return decorate


def nice_big_num_str(n, separator=','):
    '''
    Convert a number to a string with inserting separator all 3 digits.

    Note:
        With the default separator ',' one can also use the standard
        "{:,}".format(n).

    Example:
        >>> nice_big_num_str(12345678)
        '12,345,678'
        >>> nice_big_num_str(-12345)
        '-12,345'
        >>> nice_big_num_str(123456, separator=' ')
        '123 456'
        >>> nice_big_num_str(0)
        '0'
    '''
    if n < 0:
        return '-' + nice_big_num_str(-n)
    s = ''
    while n >= 1000:
        n, r = divmod(n, 1000)
        s = "%s%03d%s" % (separator, r, s)
    return "%d%s" % (n, s)


def float_to_nice_latex(x, dec=None):
    '''
    Convert a number to a nice latex representation.

    Args:
        x (float):  The number to convert.
        dec (int):  The number of digits for precision.

    Returns:
        repr (string):  The LaTeX representation.

    Example:
        >>> float_to_nice_latex(1.2345e67, 2)
        '1.23 \\\\times 10^{67}'
        >>> float_to_nice_latex(1e10)
        '10^{10}'
    '''
    s = ('%g' if dec is None else '%%.%dg' % (dec + 1)) % x
    if 'e' in s:
        # two backslashes in raw-string, because it is a regex
        # replace 'e+'
        s = re.sub(r'e\+', r' \\times 10^{', s)
        # replace the 'e' in 'e-'
        s = re.sub(r'e(?=-)', r' \\times 10^{', s)
        s += '}'
        # remove potential '1 \times '
        if s.startswith(r'1 '):
            s = s[9:]
    return s


def perm_inv(perm):
    '''
    Invert a permutation.

    Args:
        perm (array-like):  The permutation in form of an array of the
                            integers 0, 1, 2, ..., N

    Returns:
        inverse (np.ndarray):   The inverse.

    Examples:
        >>> a = np.arange(5)
        >>> ind = a.copy()
        >>> np.random.shuffle(ind)
        >>> np.all( a == a[ind][perm_inv(ind)])
        True
    '''
    return perm.argsort()


def _ndarray_periodic_distance_to(pos, center, boxsize):
    '''periodic_distance_to assuming np.ndarray's as arguments. (for speed)'''
    min_dists = np.minimum((pos - center) % boxsize,
                           (center - pos) % boxsize)
    return np.sqrt((min_dists ** 2).sum(axis=1))


def periodic_distance_to(pos, center, boxsize):
    '''
    Calculate distances in a periodic box.

    Args:
        pos (array-like):       An array of points.
        center (array-like):    The reference point.
        boxsize (float, array-like):
                                The box size. Either a float, then it the box is
                                a cube, or an array-like object, defining the
                                sidelengths for each axis individually.

    Returns:
        dist (array-like):  The distance(s) between the points.

    Examples:
        >>> from .. import units
        >>> from ..units import UnitArr
        >>> from ..environment import module_dir
        >>> units.undefine_all()
        >>> units.define_from_cfg([module_dir+'config/units.cfg'])
        reading units definitions from "units.cfg"
        >>> pos = UnitArr([[1.1,2.1,3.7], [2.8,-1.4,5.4], [7.0,3.4,-5.6]],
        ...               units='m')
        >>> ref = UnitArr([10.,120.,280.], units='cm')
        >>> periodic_distance_to(pos, ref, '5 m')
        UnitArr([1.61864141, 4.1       , 3.318132  ], units="m")
    '''
    from ..units import UnitArr

    if isinstance(boxsize, str):
        from ..units import Unit
        boxsize = Unit(boxsize)

    # go through all the unit mess...
    unit = None
    if isinstance(pos, UnitArr):
        unit = pos.units
        if isinstance(center, UnitArr):
            if not hasattr(boxsize, 'in_units_of') and pos.units != center.units:
                raise ValueError('boxsize is unitless and pos and center have '
                                 'different units. Ambiguous to interpret.')
            center = center.in_units_of(unit).view(np.ndarray)
        else:
            center = np.array(center)
        if hasattr(boxsize, 'in_units_of'):
            boxsize = boxsize.in_units_of(unit)
            if isinstance(boxsize, UnitArr):
                boxsize = boxsize.view(np.ndarray)
        pos = pos.view(np.ndarray)
    elif isinstance(center, UnitArr):
        unit = center.units
        if hasattr(boxsize, 'in_units_of'):
            boxsize = boxsize.in_units_of(unit)
            if isinstance(boxsize, UnitArr):
                boxsize = boxsize.view(np.ndarray)
        center = center.view(np.ndarray)
        pos = np.array(pos)
    elif hasattr(boxsize, 'in_units_of'):
        from ..units import _UnitClass
        unit = boxsize.units
        if isinstance(boxsize, _UnitClass):
            boxsize = float(boxsize)
        elif isinstance(boxsize, UnitArr):
            boxsize = boxsize.view(np.ndarray)
        center = np.array(center)
        pos = np.array(pos)

    r = _ndarray_periodic_distance_to(pos, center, boxsize).view(UnitArr)
    r.units = unit
    return r


def sane_slice(s, N, forward=True):
    '''
    Convert a slice into an equivalent "sane" one.

    The new slice is equivalent in the sense, that it yields the same values when
    applied. These are, however, reverted with respect to the original slice, if
    it had a negative step and `forward` is True.
    "Sane" slice means: none of the attributes of the slice is None (except for a
    backwards slice that contains the first element), 0 <= start <= stop <= N, and
    if the slice is empty the result is always slice(0,0,1).

    Args:
        s (slice):      The slice to make sane.
        N (int):        The length of the sequence to apply the slice to. (Needed
                        for negative indices and None's in the slice).
        forward (bool): Convert to forward slice.

    Returns:
        simple (slice): The sane start and stop position and the (positive) step.

    Raises:
        ValueError:     If a slice with step==0 was passed.

    Examples:
        >>> sane_slice(slice(None,2,-3), 10)
        slice(3, 10, 3)
        >>> sane_slice(slice(-3,None,None), 10)
        slice(7, 10, 1)

        Test some random parameters for the requirements:
        >>> from random import randint
        >>> for i in range(0): #10000
        ...     N = randint(0,10)
        ...     start, stop = randint(-3*N, 3*N), randint(-3*N, 3*N)
        ...     step = randint(-3*N, 3*N)
        ...     if step==0: step = None     # would raise exception
        ...     a = range(N)
        ...     s = slice(start,stop,step)
        ...     ss = sane_slice(s, N)
        ...     if not ( sorted(a[s])==a[ss] and 0<=ss.start<=ss.stop<= N
        ...                 and ss.step>0 ):
        ...         print('ERROR (forward=True):')
        ...         print(N, s, ss)
        ...         print(a[s], a[ss])
        ...     if len(a[s])==0 and not (ss.start==ss.stop==0 and ss.step==1):
        ...         print('ERROR:')
        ...         print('empty slice:', ss)
        ...     ss = sane_slice(s, N, forward=False)
        ...     if not ( a[s]==a[ss] and 0<=ss.start<=N and (
        ...                 (ss.stop is None and ss.step<0) or 0<=ss.stop<=N)):
        ...         print('ERROR (forward=False):')
        ...         print(N, s, ss)
        ...         print(a[s], a[ss])
        ...     if len(a[s])==0 and not (ss.start==ss.stop==0 and ss.step==1):
        ...         print('ERROR:')
        ...         print('empty slice:', ss)
    '''
    # get rid of None's, overly large indices, and negative indices (except -1 for
    # backward slices that go down to first element)
    start, stop, step = s.indices(N)

    # get number of steps & remaining
    n, r = divmod(stop - start, step)
    if n < 0 or (n == 0 and r == 0):
        return slice(0, 0, 1)
    if r != 0:  # it's a "stop index", not the last index
        n += 1

    if step < 0:
        if forward:
            start, stop, step = start + (n - 1) * step, start - step, -step
            stop = min(stop, N)
        else:
            stop = start + n * step
            if stop < 0:
                stop = None
    else:  # step > 0, step == 0 is not allowed
        stop = min(start + n * step, N)

    return slice(start, stop, step)


def is_consecutive(l):
    '''
    Test whether an iterable that supports indexing has consecutive elements.

    Args:
        l (iterable):   Some iterable, that supports indexing for which
                        list(l) == [l[i] for i in range(len(l))]

    Examples:
        >>> is_consecutive([1,2,3])
        True
        >>> is_consecutive([-2,-1,0,1])
        True
        >>> is_consecutive([1,3,2])
        False
        >>> is_consecutive({1:'some', 2:'dict'})
        Traceback (most recent call last):
        ...
        TypeError: Cannot check whether a dict has consecutive values!
    '''
    if isinstance(l, dict):
        raise TypeError('Cannot check whether a dict has consecutive values!')
    return np.all(np.arange(len(l)) + l[0] == list(l))


def rand_dir(dim=3):
    '''
    Create a vectors with uniform spherical distribution.

    Args:
        dim (int):      The number of dimensions the vector shall have.

    Returns:
        r (np.ndarray): A vector of shape (dim,) with length of one pointing into
                        a random direction.

    Examples:
        >>> N = 1000
        >>> for dim in [2,3,4]:
        ...     for n in range(10):
        ...         assert abs( np.linalg.norm(rand_dir(dim=dim)) - 1 ) < 1e-4
        ...     v = np.empty([N,dim])
        ...     for n in range(N):
        ...         v[n] = rand_dir(dim=dim)
        ...     assert np.linalg.norm(v.sum(axis=0))/N < 3.0/np.sqrt(N)
    '''
    if dim < 1:
        raise ValueError('Can only create vectors with at least one dimension!')
    length = 2.
    while length > 1.0 or length < 1e-4:
        r = np.random.uniform(-1., 1., size=dim)
        length = np.linalg.norm(r)
    return r / length


class ProgressBar(object):
    '''
    This function creates an iterable context manager that can be used to iterate
    over something while showing a progress bar.

    It will either iterate over the `iterable` or `length` items (that are counted
    up). While iteration happens, this function will print a rendered progress bar
    to the given `file` (defaults to stdout) and will attempt to calculate
    remaining time and more.

    The context manager creates the progress bar. When the context manager is
    entered the progress bar is already displayed. With every iteration over the
    progress bar, the iterable passed to the bar is advanced and the bar is
    updated. When the context manager exits, a newline is printed and the progress
    bar is finalized on screen.

    Note:
        No printing must happen or the progress bar will be unintentionally
        destroyed!

    Example usage:

    >>> with ProgressBar(range(100)) as pbar:
    ...     for i in pbar:
    ...         time.sleep(0.005)

    Alternatively, if no iterable is specified, one can manually update the
    progress bar through the `update()` method instead of directly iterating over
    the progress bar. The update method accepts the number of steps to increment
    the bar with:

    >>> chunks = [123, 32, 72, 201, 173, 88, 64]
    >>> with ProgressBar(length=sum(chunks)) as pbar:
    ...     for chunk in chunks:
    ...         pass # process_chunk chunk...
    ...         pbar.update(chunk)
    ...         time.sleep(0.1)

    Args:
        iterable (iterable):    An iterable to iterate over. If not provided the
                                length is required, however, it can also just be a
                                number, then the iterable will be
                                `range(iterable)`.
        length (int):           The number of items to iterate over. By default
                                the progress bar will attempt to ask the iterator
                                about its length with `len(iterable)`. Providing
                                the length, overwrites this.
                                If the iterable is None, length is required!
        label (str):            The label to show left to the progress bar.
        show_eta (bool):        Enables or disables the estimated time display.
        show_percent(bool):     Enables or disables the percentage display.
        show_iteration (bool):  Enables or disables the absolute iteration
                                display.
        item_show_func (function):
                                A function called with the current item which can
                                return a string to show the current item next to
                                the progress bar. Note that the current item can
                                be `None`!
        fill_char (str):        The character to use to show the filled part of
                                the progress bar (needs to be a single character).
        empty_char (str):       The character to use to show the non-filled part
                                of the progress bar (needs to be a single
                                character).
        bar_template (str):     The format string to use as template for the bar.
                                The parameters in it are 'label' for the label,
                                'bar' for the progress bar and 'info' for the info
                                section.
        info_sep (str):         The separator between multiple info items (eta etc.).
        width (int):            The width of the progress bar in characters, None
                                means full terminal width.
        file (file):            The file to write to.
    '''
    BEFORE_BAR = '\r\033[?25l'
    AFTER_BAR = '\033[?25h\n'

    def __init__(self, iterable=None, length=None,
                 show_eta=True, show_percent=True, show_iteration=True,
                 item_show_func=None,
                 fill_char='#', empty_char='.',
                 bar_template='%(label)s  [%(bar)s]  %(info)s',
                 info_sep='  ', file=sys.stdout, label=None,
                 width=36, eta_average_over=20):
        if iterable is None and isinstance(length, int):
            self._length = length
            self._iterable = range(length)
        else:
            if isinstance(iterable, int):
                iterable = range(iterable)
            self._length = int(length) if length else len(iterable)
            self._iterable = iter(iterable)
        self.show_eta = bool(show_eta)
        self.show_percent = bool(show_percent)
        self.show_iteration = bool(show_iteration)
        self.fill_char = str(fill_char)
        self.empty_char = str(empty_char)
        if len(self.fill_char) != 1 or len(self.empty_char) != 1:
            raise ValueError("The fill and the empty char both must have length 1.")
        self.bar_template = str(bar_template)
        self.info_sep = str(info_sep)
        self._file = file
        self.label = '' if label is None else str(label)
        self._auto_width = width is None
        self._width = 0 if width is None else width
        self._item_show_func = item_show_func
        self._eta_average_over = eta_average_over

        self._start = self._last_eta = time.time()
        self._last_line = None
        self._it_times = []
        self._it_time_its = []
        self._eta_known = False
        self._finished = False
        self._entered = False
        self._it = 0
        self._current_item = None

    def __enter__(self):
        self._entered = True
        self._start = time.time()
        self._it = 0
        self.render()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self._render_finish()

    def __iter__(self):
        if not self._entered:
            raise RuntimeError('You need to use progress bars in a with block.')
        self.render()
        return self

    def __next__(self):
        try:
            rv = next(self._iterable)
            self._current_item = rv
        except StopIteration:
            self._finish()
            self.render()
            raise StopIteration()
        else:
            self.update(1)
            return rv

    def __len__(self):
        return self._length

    @property
    def auto_width(self):
        return self._auto_width

    @auto_width.setter
    def auto_width(self, value):
        self._auto_width = bool(value)

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        if value is None:
            self._auto_width = True
        else:
            self._width = max(0, int(value))

    @property
    def iteration(self):
        '''The iteration being processed.'''
        return self._it

    @property
    def pct(self):
        '''The percent done (as fraction, i.e. finished is 1.0).'''
        return min(float(self._it) / self._length, 1.0)

    @property
    def time_per_iteration(self):
        if len(self._it_times) < 2:
            return 0.0
        return np.ptp(self._it_times) / float(np.ptp(self._it_time_its))

    @property
    def eta(self):
        if not self._finished:
            return self.time_per_iteration * (self._length - self._it)
        return 0.0

    def format_iteration(self):
        '''The iteration formatted to a string.'''
        return '%s/%s' % (self._it, self._length)

    def format_pct(self):
        '''The percent done formatted to a string.'''
        return '%3d%%' % int(self.pct * 100.)

    def format_eta(self):
        '''The ETA formatted to a string (if unknown return '').'''
        if self._eta_known:
            t = int(self.eta + 1)
            seconds = t % 60
            t /= 60
            minutes = t % 60
            t /= 60
            hours = t % 24
            t /= 24
            if t > 0:
                days = t
                return '%dd %d:%02d:%02d' % (days, hours, minutes, seconds)
            else:
                return '%d:%02d:%02d' % (hours, minutes, seconds)
        return ''

    def format_progress_line(self):
        '''Create the progress bar string (for printing use `render`).'''
        info_bits = []

        bar_length = int(self.pct * self._width)
        bar = self.fill_char * bar_length
        bar += self.empty_char * (self._width - bar_length)

        if self.show_iteration:
            info_bits.append(self.format_iteration())
        if self.show_percent:
            info_bits.append(self.format_pct())
        if self.show_eta and self._eta_known and not self._finished:
            info_bits.append(self.format_eta())
        if self._item_show_func is not None:
            item_info = self._item_show_func(self._current_item)
            if item_info is not None:
                info_bits.append(str(item_info))

        return (self.bar_template % {
            'label': self.label,
            'bar': bar,
            'info': self.info_sep.join(info_bits)
        }).rstrip()

    def render(self):
        '''Render the progress bar.'''
        buf = []

        if self.auto_width:
            old_width = self._width
            self._width = 0
            clutter_length = term_len(self.format_progress_line())
            new_width = max(0, get_terminal_size()[0] - clutter_length)
            if new_width < old_width:
                buf.append(self.BEFORE_BAR)
                buf.append(' ' * old_width)
            self._width = new_width

        clear_width = self._width

        buf.append(self.BEFORE_BAR)
        line = self.format_progress_line()
        line_len = term_len(line)
        buf.append(line)

        buf.append(' ' * (clear_width - line_len))
        line = ''.join(buf)

        if line != self._last_line:
            self._last_line = line
            self._file.write(line)
            self._file.flush()

    def _render_finish(self):
        self._file.write(self.AFTER_BAR)
        self._file.flush()

    def _finish(self):
        self._eta_known = False
        self._current_item = None
        self._finished = True

    def update(self, n_steps):
        '''
        Advance the progress bar by the given number of steps and re-render it.

        Args:
            n_steps (int):  The number of steps to advance.
        '''
        self._make_step(n_steps)
        self.render()

    def _make_step(self, n_steps):
        self._it += n_steps
        if self._it >= self._length:
            self._finished = True

        t = time.time()
        if (t - self._last_eta) < 1.0:
            # not sufficiently good estimate
            return

        self._last_eta = t
        self._it_times = self._it_times[-self._eta_average_over:] + [t]
        self._it_time_its = self._it_time_its[-self._eta_average_over:] + \
                            [self._it]
        self._eta_known = True


def time_from_color_str(time, units='s'):
    '''
    TODO!

    Examples:
        >>> time_from_color_str('6:12')
        UnitArr(372.0, units="s")
        >>> time_from_color_str('1:20:00', units='h')
        UnitArr(1.3333333333333333, units="h")
    '''
    from ..units import UnitArr
    nums = time.split(':')
    t = 0
    for val in nums:
        t = 60 * t + float(val)
    return UnitArr(t, 's').in_units_of(units)


def time_to_nice_str(time, colon=False, opt_hour=False, prec=None, days=None):
    '''
    Convert a number of seconds to a nice string.

    Args:
        time (UnitScalar):  The time unit scalar to format into a nice string.
        colon (bool):       Use colons to seperate hours, minutes, and seconds,
                            not letters.
        opt_hour (bool):    Do not print the number of hours (and days) if they
                            are zero.
        prec (int):         The precision of the seconds. If not specified, it
                            will either just be integer precision for integral
                            values of secs, or it defaults to a precision of
                            three otherwise.
        days (bool):        Whether to print the number of days or just break it
                            down to hours. When not specified / None, the number
                            of days is only printed, it the duration (`secs`)
                            exceeds 24h.

    Returns:
        repr (str):         A human readable string representation of the time in
                            (days,) hours, minutes, and seconds.

    Examples:
        >>> time_to_nice_str(62*60+3)
        '1h 2m 3s'
        >>> time_to_nice_str( '73 min' )
        '1h 13m 0s'
        >>> time_to_nice_str( '73 min', colon=True )
        '01:13:00'
        >>> time_to_nice_str(int(123e4), prec=1)
        '14d 5h 40m 0.0s'
        >>> time_to_nice_str( time_from_color_str('3:11'), colon=True, opt_hour=True )
        '3:11'
    '''
    from ..units import UnitScalar
    time = UnitScalar(time, 's')
    if prec is None:
        if abs(time) > 1 and abs(time - int(time)) < 1e-12:
            prec = 0
    return sec_to_nice_str(float(time), colon=colon, opt_hour=opt_hour, prec=prec, days=days)


def sec_to_nice_str(secs, colon=False, opt_hour=False, prec=None, days=None):
    '''
    Convert a number of seconds to a nice string.

    Args:
        secs (int, float):  The number of seconds to format into a nice string.
        colon (bool):       Use colons to seperate hours, minutes, and seconds,
                            not letters.
        opt_hour (bool):    Do not print the number of hours (and days) if they
                            are zero.
        prec (int):         The precision of the seconds. If not specified, it
                            will either just be integer precision for integral
                            values of secs, or it defaults to a precision of
                            three otherwise.
        days (bool):        Whether to print the number of days or just break it
                            down to hours. When not specified / None, the number
                            of days is only printed, it the duration (`secs`)
                            exceeds 24h.

    Returns:
        repr (str):         A human readable string representation of the time in
                            (days,) hours, minutes, and seconds.

    Examples:
        >>> sec_to_nice_str(12345)
        '3h 25m 45s'
        >>> sec_to_nice_str(12345, colon=True)
        '03:25:45'
        >>> sec_to_nice_str(123e4)
        '14d 5h 40m 0.000s'
        >>> sec_to_nice_str(123e4, colon=True)
        '14d 05:40:00.000'
        >>> sec_to_nice_str(int(123e4))
        '14d 5h 40m 0s'
        >>> sec_to_nice_str(int(123e4), days=False)
        '341h 40m 0s'
        >>> sec_to_nice_str(int(123e4), colon=True)
        '14d 05:40:00'
        >>> sec_to_nice_str(int(123e4), colon=True, days=False)
        '341:40:00'
        >>> sec_to_nice_str(123, opt_hour=True)
        '2m 3s'
        >>> sec_to_nice_str(123, colon=True, opt_hour=True)
        '2:03'
        >>> sec_to_nice_str(180, days=True)
        '0d 0h 3m 0s'
        >>> sec_to_nice_str(123.456)
        '0h 2m 3.456s'
        >>> sec_to_nice_str(123.456, prec=1)
        '0h 2m 3.5s'
        >>> sec_to_nice_str(1.23e-7)
        '0h 0m 0.000000123s'
        >>> sec_to_nice_str(1.23e-12)
        '0h 0m 0.000000000s'
    '''
    if secs < 0:
        raise ValueError("Negative durations are not supported!")
    if not (prec is None or (isinstance(prec, int) and prec >= 0)):
        raise ValueError("Precision needs to be a positive integer!")

    m, s = divmod(secs, 60)
    m = int(m)
    h, m = divmod(m, 60)

    if isinstance(secs, numbers.Integral):
        tformat = '%dh %dm %ds'
        prec = 0
    else:
        if prec is None:
            prec = 3 if secs > 1 else min(9, 3 - np.log10(secs))
        tformat = '%%dh %%dm %%.%dfs' % prec
    if colon:
        tformat = tformat[:-1].replace('h ', ':').replace('m ', ':')
        tformat = tformat.replace('%d', '%02d')
        tformat = tformat.replace('%.', '%%0%d.' % (2 + (prec > 0) + prec))

    if days or (days is None and h >= 24):
        d, h = divmod(h, 24)
        string = ('%dd ' + tformat) % (d, h, m, s)
    else:
        d = None
        string = tformat % (h, m, s)

    if opt_hour and (d is None or d == 0) and h == 0:
        if colon:
            string = string[string.find(':') + 1:]
        else:
            string = string[string.find(' ') + 1:]
        string = string.lstrip('0')
        if string[0] == ':':
            string = '0' + string

    return string


def weighted_percentile(qty, perc, weights=None, values_sorted=False,
                        ignore_nan=False, new_style=False):
    '''
    Very close to np.percentile, but supports weights.

    Args:
        qty (array-like):       The data to calculate the percentiles for.
        perc (array-like):      The percentiles to calculate. Must be in the range
                                [0,100], of course.
        weights (array-like):   The weights (or counts) for the values in `qty`.
                                Needs to be the same length as `qty`, with a
                                one-to-one correspondence to the values.
        values_sorted (bool):   If True, it will be assumed that `qty` already is
                                a sorted array.
        ignore_nan (bool):      Whether to ignore values of NaN in both `qty` and
                                `weights`, i.e. to cut them out in advance.
        new_style (bool):       If False, the output will be consistent with
                                np.percentile; if True, the interpolation is
                                slightly different: the 33-percentile of [0,0.5,1]
                                no longer is 0.333..., but 0.25 (simply the mean
                                of the two first elements).

    Returns:
        perc (array-like):      The percentiles.

    Examples:
        >>> weighted_percentile( [0,0.5,1], [100./3], new_style=False)
        array([0.33333333])
        >>> np.percentile( [0,0.5,1], [100./3] )[0] # doctest: +ELLIPSIS
        0.333333...
        >>> weighted_percentile( [0,0.5,1], [100./3], new_style=True)
        array([0.25])
        >>> weighted_percentile( [4,5,1,10,6,2,3], [25,50,75],
        ...                      weights=[12,8,7,10,9,3,4])
        array([3.328125  , 4.575     , 5.80882353])
        >>> weighted_percentile( [3,4,2,7,4,1], [0,54.21,100],
        ...                      weights=[1,2,1,2,1,2])
        array([1.        , 3.86313333, 7.        ])
    '''
    if not isinstance(qty, np.ndarray):
        qty = np.array(qty)
    perc = np.array(perc, dtype=float) / 100.
    if np.any(perc < 0) or np.any(perc > 100):
        raise ValueError('Percentiles need to be be in [0,100]!')
    if weights is None:
        weights = np.ones(len(qty))
    weights = np.array(weights, dtype=float)
    if len(weights) != len(qty):
        raise ValueError('Weights need to be the same length as `qty`!')

    if ignore_nan:
        not_nan = ~(np.isnan(qty) | np.isnan(weights))
        qty = qty[not_nan]
        weights = weights[not_nan]

    if not values_sorted:
        sorter = np.argsort(qty)
        qty = qty[sorter]
        weights = weights[sorter]

    weighted_quantiles = np.cumsum(weights) - 0.5 * weights
    if new_style:
        weighted_quantiles /= np.sum(weights)
    else:
        # consistent with np.percentile
        weighted_quantiles -= weighted_quantiles[0]
        weighted_quantiles /= weighted_quantiles[-1]
    return np.interp(perc, weighted_quantiles, qty)




'''
Some general (low-level) functions.

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(utils)
    TestResults(failed=0, attempted=62)
    >>> doctest.testmod(safe_eval)
    TestResults(failed=0, attempted=11)
    >>> doctest.testmod(term)
    TestResults(failed=0, attempted=0)
    >>> doctest.testmod(geo)
    TestResults(failed=0, attempted=20)
'''
from .utils import *
from .safe_eval import *
from .term import *
from .geo import *




'''
Some general high-level functions.

Example:
'''
__all__ = ['read_info_file', 'prepare_zoom', 'fill_star_from_info',
           'read_traced_gas', 'fill_gas_from_traced',
           'fill_derived_gas_trace_qty', 'read_pyigm_COS_halos_EWs']

from .snapshot import *
from .units import *
from .utils import *
from .analysis import *
from .transformation import *
from . import environment
import re
import sys
import os
import numpy as np


def read_info_file(filename):
    '''
    Read in the contents of an info file as produced by gtrace into a dictionary.

    It is assumed, that there is exactly one colon per line, which separates the
    name (the part before the colon) and the value (the part after the colon). For
    the value part the following is assumed: if there are no brackets ('[') it is
    a float; if there is one pair of brackets, it is a float with units, which are
    given within the brackets; otherwise the value is simply stored as a string.

    Args:
        filename (str):     The path to the info file.

    Returns:
        info (dict):        A dictionary containing all the entries from the file.
    '''
    info = {}
    with open(os.path.expanduser(filename), 'r') as finfo:
        for line in finfo:
            try:
                cols = line.split(':', 1)
                if len(cols) > 1:           # otherwise no value given, avoid ValueError
                    name = cols[0].strip()
                    value = cols[1].strip()
                    blocks = re.findall('\[(.*?)\]', value)
                    if len(blocks) == 0:
                        value = float(value) if value != 'None' else None
                    elif len(blocks) == 1:
                        value = UnitArr(float(value.rsplit('[')[0].strip()),
                                        units=blocks[0])
                    elif len(blocks) == 2:
                        value = UnitArr([float(s.strip(', ')) for s in blocks[0].split()],
                                        units=None if blocks[1] == 'None' else blocks[1])
                    if name in info:
                        print('WARNING: "%s" occures ' % name + \
                              'multiple times in info file ' + \
                              '"%s"! First one used.' % filename, file=sys.stderr)
                    info[name] = value
            except ValueError as e:
                continue    # ignore error continue loading

    return info


def prepare_zoom(s, mode='auto', info='deduce', shrink_on='stars',
                 linking_length=None, linking_vel='200 km/s', ret_FoF=False,
                 sph_overlap_mask=False, gal_R200=0.10, star_form='deduce',
                 gas_trace='deduce', to_physical=True, load_double_prec=False,
                 fill_undefined_nan=True, gas_traced_blocks='all',
                 gas_traced_dervied_blocks=None, center_on_BH=False, **kwargs):
    '''
    A convenience function to load a snapshot from a zoomed-in simulation that is
    not yet centered or oriented.

    Args:
        s (str, Snap):      The snapshot of a zoomed-in simulation to prepare.
                            Either as an already loaded snapshot or a path to the
                            snapshot.
        mode (str):         The mode in which to prepare the snapshot. You can
                            choose from:
                                * 'auto':   try 'info', if it does not work
                                            fallback to 'FoF'
                                * 'info':   read the info file (from `info`) and
                                            take the center and the reduced
                                            inertia tensor (for orientation) with
                                            fallback to angular momentum from this
                                            file
                                * 'ssc':    center the snapshot using a
                                            shrinking sphere method (shrinking on
                                            the sub-snapshot specified by
                                            `shrink_on`, see below)
                                            ATTENTION: can easily center on a
                                            non-resolved galaxy!
                                * 'FoF':    Do a FoF search on all particles and
                                            generate a halo catalogue for all
                                            those halos that have at most 1% mass
                                            from "lowres" particles. Then find the
                                            most massive FoF group of particles
                                            defined by `shrink_on` therein and
                                            finally perform a shrinking sphere on
                                            them to find the center. If
                                            `shrink_on` is 'all' or 'highres',
                                            skip the second FoF search and do the
                                            "ssc" on the most massive FoF group of
                                            the first step. The linking length for
                                            (both) the FoF finder is given by
                                            `linking_length`.
        info (str, dict):   Path to info file or the dictionary as returned from
                            `read_info_file`.
                            However, if set to 'deduce', it is tried to deduce the
                            path to the info file from the snapshot filename: it
                            is assumed to be in a subfolder 'trace' that is in the
                            same directory as the snapshot and named
                            `info_%03d.txt`, where the `%03d` is filled with the
                            snapshot number. The latter is taken as the last three
                            characters of the first dot / the end of the filename.
        shrink_on (str, list):
                            Define the part of the snapshot to perform the
                            shrinking sphere on, if there is no info file found or
                            info==None. It can be 'all' (use the entire
                            (sub-)snapshot `s`), a family name, or a list of
                            particle types (e.g. [0,1,4]).
        linking_length (UnitScalar):
                            The linking length used for the FoF finder in mode
                            "FoF" (if None, it defaults to
                            ( 1500.*rho_crit / median(mass) )^(-1/3) as defined in
                            `generate_FoF_catalogue`).
        linking_vel (UnitScalar):
                            The linking velocity used for the FoF finder in mode
                            "FoF". Only used for defining the galaxy, not for the
                            halo.
        ret_FoF (bool):     Also return the FoF group catalogue created (if in
                            mode "FoF").
        sph_overlap_mask (bool):
                            Whether to mask all particles that overlap into the
                            halo region (that is also include SPH particles that
                            are outside the virial radius, but their smoothing
                            length reaches into it).
        gal_R200 (float):   The radius to define the galaxy. Everything within
                            <gal_R200>*R200 will be defined as the galaxy.
        star_form (str):    Path to the star formation file as written by the
                            program `gtrace`.
                            If set to 'deduce', its path is tried to build from
                            the snapshot filename by taking its directory and
                            adding '/trace/star_form.ascii'.
        gas_trace (str):    Path to the tracing file as written by the program
                            `gtracegas`.
                            If set to 'deduce', its path is tried to build from
                            the snapshot filename by taking its directory and
                            looking for files '/../*gastrace*'.
        to_physical (bool): Whether to convert the snapshot to physical units.
        load_double_prec (bool):
                            Force to load all blocks in double precision.
                            Equivalent with setting the snapshots attribute.
        fill_undefined_nan (bool):
                            Fill star formation info for IDs not listed in the
                            `star_form` file. For more info see
                            `fill_star_from_info`.
        gas_traced_blocks (iterable, str):
                            The blocks to add with `fill_gas_from_traced`.
        gas_traced_dervied_blocks (bool):
                            Passed to `fill_gas_from_traced` as `add_derived`.
                            Defaults to `gas_traced_blocks=='all'` if None.
        center_on_BH (bool):
                    If True, positions and velocities are centered on
                    the central supermassive black hole (if present).
                    If there are multiple supermassive black holes
                    within the central kpc of the galaxy, their center
                    of mass is used as the center.
        kwargs:             Passed to `generate_FoF_catalogue` in `mode='FoF'`.

    Returns:
        s (Snap):           The prepared snapshot.
        halo (SubSnap):     The cut halo of the found structure.
        gal (SubSnap):      The central galaxy of the halo as defined by
                            `gal_R200`.
    '''

    def get_shrink_on_sub(snap, shrink_on):
        if isinstance(shrink_on, str):
            shrink_on = str(shrink_on)
            if shrink_on == 'all':
                return snap
            else:
                return getattr(s, shrink_on)
        elif isinstance(shrink_on, list):
            return s[shrink_on]
        else:
            raise ValueError('`shrink_on` must be a family name or a list ' + \
                             'of particle types, but was: %s' % (shrink_on,))

    def FoF_exclude(h, s, threshold=1e-2):
        M_lowres = h.lowres_mass
        M = h.mass
        return M_lowres / M > threshold

    if 'gastrace' in kwargs:
        raise ValueError("You passed 'gastrace'. Did you mean 'gas_trace'?")
    if 'starform' in kwargs:
        raise ValueError("You passed 'starform'. Did you mean 'star_form'?")

    if isinstance(s, str):
        s = Snapshot(s, load_double_prec=load_double_prec)
    gal_R200 = float(gal_R200)
    if environment.verbose >= environment.VERBOSE_TACITURN:
        print('prepare zoomed-in', s)

    # read info file (if required)
    if mode in ['auto', 'info']:
        if info == 'deduce':
            try:
                snap = int(os.path.basename(s.filename).split('.')[0][-3:])
                info = os.path.dirname(s.filename) + '/trace/info_%03d.txt' % snap
            except:
                print('WARNING: could not deduce the path to ' + \
                      'the trace file!', file=sys.stderr)
                info = None
        if isinstance(info, str):
            info = os.path.expanduser(info)
            if not os.path.exists(info):
                print('WARNING: There is no info file named ' + \
                      '"%s"' % info, file=sys.stderr)
                info = None
            else:
                if environment.verbose >= environment.VERBOSE_TACITURN:
                    print('read info file from:', info)
                info = read_info_file(info)
        if info is None:
            if mode == 'auto':
                mode = 'FoF'
            else:
                raise IOError('Could not read/find the info file!')
        else:
            if mode == 'auto':
                mode = 'info'

    if to_physical:
        s.to_physical_units()

    # find center
    if mode == 'info':
        center = info['center']
    elif mode in ['ssc', 'FoF']:
        if mode == 'FoF':
            halos = generate_FoF_catalogue(
                s,
                l=linking_length,
                exclude=FoF_exclude,
                calc=['mass', 'lowres_mass'],
                max_halos=10,
                progressbar=False,
                **kwargs
            )
            if shrink_on not in ['all', 'highres']:
                galaxies = generate_FoF_catalogue(
                    get_shrink_on_sub(s, shrink_on),
                    l=linking_length,
                    dvmax=linking_vel,
                    calc=['mass', 'com'],
                    max_halos=10,
                    progressbar=False,
                    **kwargs
                )
                # The most massive galaxy does not have to be in a halo with litle
                # low resolution elements! Find the most massive galaxy living in a
                # "resolved" halo:
                galaxy = None
                for gal in galaxies:
                    # since the same linking lengths were used for the halos and
                    # the galaxies (rethink that!), a galaxy in a halo is entirely
                    # in that halo or not at all
                    gal_ID_set = set(gal.IDs)
                    for h in halos:
                        if len(gal_ID_set - set(h.IDs)) == 0:
                            galaxy = gal
                            break
                    del h, gal_ID_set
                    if galaxy is not None:
                        break
                del galaxies
                if galaxy is None:
                    shrink_on = None
                else:
                    shrink_on = s[galaxy]
            else:
                shrink_on = s[halos[0]]
        elif mode == 'ssc':
            shrink_on = get_shrink_on_sub(s, shrink_on)
        if shrink_on is not None and len(shrink_on) > 0:
            com = center_of_mass(s)
            R = np.max(periodic_distance_to(s['pos'], com, s.boxsize))
            center = shrinking_sphere(shrink_on, com, R)
        else:
            center = None
    else:
        raise ValueError('Unkown mode "%s"!' % mode)

    # center in space
    if center is None:
        if environment.verbose >= environment.VERBOSE_TACITURN:
            print('no center found -- do not center')
    else:
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('center at:', center)
        Translation(-center).apply(s)
        # center the velocities
        vel_center = mass_weighted_mean(s[s['r'] < '1 kpc'], 'vel')
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('center velocities at:', vel_center)
        s['vel'] -= vel_center

    # center galaxy on central supermassive black hole(s)
    if center_on_BH:
        bh_search_rad_kpc = 1.
        search_ball=s[BallMask(str(bh_search_rad_kpc)+' kpc')]
        if search_ball.bh["mass"].size==0: #No BH
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print("No black holes found, center stays the same.")
        else:
            if search_ball.bh["mass"].size > 1: #>1 BHs
                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print("WARNING: Multiple black holes within the central kpc, centering on their center of mass")
                #selecting the most massive black hole
                bhpos = np.average(search_ball.bh["pos"], axis=0,
                                   weights=search_ball.bh["mass"] )
                bhvel = np.average(search_ball.bh["vel"], axis=0,
                                   weights=search_ball.bh["mass"] )
                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print("Center of mass: " + str(bhpos))
                    print("Velocity of center of mass: " + str(bhvel))
            else:  #1 BH
                bhpos = search_ball.bh["pos"][0]
                bhvel = search_ball.bh["vel"][0]
                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print("Central black hole position: " + str(bhpos))
                    print("Central black hole velocity: " + str(bhvel))
            s["pos"] = s["pos"] - bhpos
            s["vel"] = s["vel"] - bhvel


    # cut the halo (<R200)
    if mode == 'info':
        R200 = info['R200']
        M200 = info['M200']
    else:
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('derive virial information')
        R200, M200 = virial_info(s)
    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('R200:', R200)
        print('M200:', M200)
    halo = s[BallMask(R200, sph_overlap=sph_overlap_mask)]

    # orientate at the reduced inertia tensor of the baryons wihtin 10 kpc
    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('orientate', end=' ')
    if mode == 'info':
        if 'I_red(gal)' in info:
            redI = info['I_red(gal)']
            if redI is not None:
                redI = redI.reshape((3, 3))
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('at the galactic red. inertia tensor from info file')
            if environment.verbose >= environment.VERBOSE_TALKY:
                print(redI)
            mode, qty = 'red I', redI
        else:
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('at angular momentum of the galaxtic baryons from info file:')
            mode, qty = 'vec', info['L_baryons']
        orientate_at(s, mode, qty=qty, total=True)
    else:
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('at red. inertia tensor of the baryons within %.3f*R200' % gal_R200)
        orientate_at(shrink_on[BallMask(gal_R200*R200, sph_overlap=False)],
                     'red I',
                     total=True
                     )

    # cut the inner part as the galaxy
    gal = s[BallMask(gal_R200 * R200, sph_overlap=sph_overlap_mask)]
    Ms = gal.stars['mass'].sum()
    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('M*:  ', Ms)

    if len(gal) == 0:
        gal = None
    if len(halo) == 0:
        halo = None

    if star_form == 'deduce':
        try:
            star_form = os.path.dirname(s.filename) + '/trace/star_form.ascii'
        except:
            print('WARNING: could not deduce the path to the ' + \
                  'star formation file!', file=sys.stderr)
            star_form = None
    if isinstance(star_form, str):
        star_form = os.path.expanduser(star_form)
        if not os.path.exists(star_form):
            print('WARNING: There is no star formation file ' + \
                  'named "%s"' % star_form, file=sys.stderr)
            star_form = None
        else:
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('read star formation file from:', star_form)
            fill_star_from_info(s, star_form,
                                fill_undefined_nan=fill_undefined_nan)

    if gas_trace == 'deduce':
        try:
            directory = os.path.dirname(s.filename) + '/../'
            candidates = []
            for fname in os.listdir(directory):
                if fname.startswith('gastrace'):
                    candidates.append(fname)
            if len(candidates) == 1:
                gas_trace = directory + candidates[0]
            else:
                raise RuntimeError('too many candidates!')
        except:
            print('WARNING: could not deduce the path to the ' + \
                  'gas tracing file!', file=sys.stderr)
            gas_trace = None
    if isinstance(gas_trace, str):
        gas_trace = os.path.expanduser(gas_trace)
        if not os.path.exists(gas_trace):
            print('WARNING: There is no gas trace file named ' + \
                  '"%s"' % gas_trace, file=sys.stderr)
            gas_trace = None
        else:
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('read gas trace file from:', gas_trace)
            if gas_traced_dervied_blocks is None:
                gas_traced_dervied_blocks = (gas_traced_blocks == 'all')
            fill_gas_from_traced(s, gas_trace,
                                 add_blocks=gas_traced_blocks,
                                 add_derived=gas_traced_dervied_blocks)

    if mode == 'FoF' and ret_FoF:
        return s, halo, gal, halos
    else:
        return s, halo, gal


def fill_star_from_info(snap, fname, fill_undefined_nan=True, dtypes=None,
                        units=None):
    '''
    Read the formation radius rform and rform/R200(aform) from the star_form.ascii
    file and create the new blocks "rform" and "rR200form".

    Note:
        The set of the star particle IDs in the star formation information file
        must exactly be the set of star particle IDs in the (root) snapshot.

    Args:
        snap (Snap):    The snapshot to fill with the data (has to be the one at
                        z=0 of the simulation used to create the star_form.ascii).
        fname (str):    The path to the star_form.ascii file.
        fill_undefined_nan (bool):
                        Fill data with NaN for stars that are not listed in the
                        formation file.
        dtypes (np.dtypes):
                        The named(!) fields in the formation file. One of the has
                        to be a (integer type) field called "ID" for matching the
                        data to the snapshot.
        units (dict):   Specifying the units of the fields. Not every field has to
                        be sopecified here.
                        Defaults to: `{'aform':'a_form', 'rform':'kpc'}`.

    Raises:
        RuntimeError:   If the IDs are not unique or they do not match (except the
                        cases where `fill_undefined_nan` applies).
    '''
    stars = snap.root.stars

    if environment.verbose >= environment.VERBOSE_TACITURN:
        print('reading the star formation information from %s...' % fname)
    # prepare the type of data
    if dtypes is None:
        dtypes = [('ID', np.uint64), ('aform', float), ('rform', float),
                  ('rR200form', float), ('Zform', float)]
    dtypes = np.dtype(dtypes)
    if 'ID' not in dtypes.fields:
        raise ValueError('The `dtypes` need to have a field "ID"!')
    if units is None:
        units = {'aform': 'a_form', 'rform': 'kpc'}
    # load the data
    SFI = np.loadtxt(fname, skiprows=1, dtype=dtypes)

    if environment.verbose >= environment.VERBOSE_TALKY:
        print('testing if the IDs match the (root) snapshot...')
    SFI_IDs = SFI['ID']
    # test uniqueness
    if not stars.IDs_unique():
        raise RuntimeError('Stellar IDs in the snapshot are not unique!')
    if len(np.unique(SFI_IDs)) != len(SFI_IDs):
        raise RuntimeError('IDs in the star formation file are not unique!')
    # there might be too many or not enough IDs in the file
    if len(np.setdiff1d(stars['ID'], SFI_IDs, assume_unique=True)):
        missing = np.setdiff1d(stars['ID'], SFI_IDs, assume_unique=True)
        if fill_undefined_nan:
            print('WARNING: There are %d stellar IDs missing ' % len(missing) + \
                  'in the formation file! Fill them with NaN\'s.')
            add = np.array([(ID,) + (np.NaN,) * (len(dtypes) - 1) for ID in missing],
                           dtype=dtypes)
            SFI = np.concatenate((SFI, add))
            del add
            SFI_IDs = SFI['ID']
        else:
            raise RuntimeError('Some stars do not have a match in the ' + \
                               'formation file "%s" (missing: %d)!' % (fname,
                                                                       len(np.setdiff1d(stars['ID'], SFI_IDs,
                                                                                        assume_unique=True)))
                               )
    if len(np.setdiff1d(SFI_IDs, stars['ID'], assume_unique=True)):
        raise RuntimeError('Some formation file IDs do not have a match in ' + \
                           'the snapshot (missing: %d)!' % (
                               len(np.setdiff1d(SFI_IDs, stars['ID'],
                                                assume_unique=True)))
                           )

    # adding the data as blocks
    sfiididx = np.argsort(SFI_IDs)
    sididx = np.argsort(stars['ID'])
    for name in dtypes.names:
        if name == 'ID':
            continue
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('adding the new block "%s" (units:%s, type:%s)...' % (
                name, units.get(name, None), dtypes[name]))
        stars[name] = UnitArr(np.empty(len(stars)),
                              dtype=dtypes[name],
                              units=units.get(name, None))
        stars[name][sididx] = SFI[name][sfiididx]


def read_traced_gas(filename, types=None):
    '''
    Read the gas tracing statistics from a gtracegas output.

    The data also gets tagged by type:
        1:  gas in region
        2:  gas out of region
        3:  stars that formed in region
        4:  stars that formed outside region
    and the number of full cycles (leaving region + re-entering).

    TODO:
        * fill the stars that have been traced gas!

    The following is tracked at the different events:
        (re-)entering the region:       [a, mass, metals, j_z, T]
        the leave the region:           [a, mass, metals, j_z, T, vel]
        turning into a star:            [a, mass, metals, j_z]
        being out of the region, the following is updated:
                                        [(a, r_max), (a, z_max)]
    where a is the scale factor of when the particle reached the maximum radius or
    height, respectively.

    Args:
        filename (str):     The filename to read from.
        types (iterable):   The types of particles (see above!) to filter for.
                            Default: all.

    Returns:
        tr (dict):          A dictionary with the trace data for the individual
                            particles (of types ask for) indexed by ID. The actual
                            data is:
                                [ [type, n],
                                  np.ndarray of first enter event,
                                  np.ndarray leaving event,          \
                                  np.ndarray of traced being out,     >  n times
                                  np.ndarray of re-entering,         /
                                  events of leaving + out,          \  depending
                                  forming a star, or                 > on type or
                                  leaving + out + forming a stars,  / just nothing
                                ]
                            where n is the number of full recylces.
    '''
    filename = os.path.expanduser(filename)
    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('read gas trace file "%s"...' % filename)
        sys.stdout.flush()
    import pickle as pickle
    with open(filename, 'rb') as f:
        tr = pickle.load(f)

    # sort type
    if types is None or types == 'all':
        types = {1, 2, 3, 4}
    else:
        if isinstance(types, int):
            types = [types]
        types = set(types)
    if environment.verbose >= environment.VERBOSE_TALKY:
        print('restructure by type...')
        sys.stdout.flush()
    # structure the data into sub-lists
    # (re-)enter:   5 elements
    # leave:        6 elements
    # star form:    4 elements
    # being out:    4 elements
    # -> full recycle:  6+4+5 = 15
    # -> in region:     5 + n*15                =  5 + n*15
    #    out:           5 + n*15 + 6 + 4        = 15 + n*15
    #    SF in region:  5 + n*15 + 4            =  9 + n*15
    #    SF outside:    5 + n*15 + 6 + 4 + 4    = 19 + n*15
    t_to_type = {5: 1, 15: 2, 9: 3, 19: 4}
    for ID in list(tr.keys()):
        e = tr[ID]
        t = len(e) % 15
        if t in [0, 4]: t += 15
        tt = t_to_type[t]
        if tt not in types:
            print('ERROR: could not process particle ID %d' % ID, file=sys.stderr)
            print('       skip: %s' % e, file=sys.stderr)
            del tr[ID]
            continue
        sub = e[5: {5: None, 15: -10, 9: -4, 19: -14}[t]]
        # assert len(sub) % 15 == 0
        new = [[tt, len(sub) / 15]]
        new += [e[:5]]
        for re in np.array(sub).reshape(((len(e) - t) / 15, 15)):
            new += [re[:6]]
            new += [re[6:10]]
            new += [re[10:]]
        if t == 5:  # gas in region (with possible recycles)
            pass
        elif t == 9:  # turned in the region into a star
            new += [e[-4:]]
        elif t == 15:  # left region, but is still a gas particle
            new += [e[-10:-4], e[-4:]]
        elif t == 19:  # left region and turned into a star
            new += [e[-14:-8], e[-8:-4], e[-4]]
        else:
            raise RuntimeError('Structure in "%s" ' % filename + \
                               'is not as expected!')
        tr[ID] = new

    return tr


def fill_gas_from_traced(snap, data, add_blocks='all', add_derived=True,
                         units=None, invalid=0.0):
    '''
    Fill some information from the gas trace file into the snapshot as blocks.

    This function is using data from a gas trace file (as produced by `gtracegas`)
    to create new gas blocks with particle properties at in- and outflow times as
    well as additional information of the time being ejected.

    Among the new blocks are:
      "trace_type", "num_recycled",
      "infall_a", "infall_time",
      "mass_at_infall", "metals_at_infall", "jz_at_infall", "T_at_infall",
      "ejection_a", "ejection_time",
      "mass_at_ejection", "metals_at_ejection", "jz_at_ejection", "T_at_ejection",
      "cycle_r_max_at", "cycle_r_max", "cycle_z_max_at", "cycle_z_max",
    For the addtional derived blocks see `fill_derived_gas_trace_qty`.

    Note:
        The set of the star particle IDs in the star formation information file
        must exactly be the set of star particle IDs in the (root) snapshot.

    Args:
        snap (Snap):        The snapshot to fill with the data (has to be the one
                            at z=0 of the simulation used to create the trace
                            file).
        data (str, dict):   The path to the gas trace file or the alread read-in
                            data.
        add_blocks (iterable, str):
                            A list of the block names to actually add to the gas
                            sub-snapshot, or just 'all', which then adds all of
                            them. If `add_derived` it True, it will be set to 'all'
                            anyway, since they might be needed afterwards.
        add_derived (bool): Whether to also add blocks that can be derived from
                            the trace data stored in blocks. The function
                            `fill_derived_gas_trace_qty` is used for that.
        units (dict):       The units to use for masses, lengths, etc..
                            Defaults to: {'TIME':'a_form', 'MASS':'Msol',
                                          'TEMP':'K', 'ANGMOM':None, 'POS':'kpc'}
        invalid (float/NaN/inf):
                            The value to fill invalid entries with. Such entries
                            are properties of cycles that did not happen for a
                            given particle.
    '''
    if add_derived:
        add_blocks = 'all'
    if add_blocks == 'all':
        add_blocks = ["trace_type", "num_recycled",
                      "infall_a", "infall_time",
                      "mass_at_infall", "metals_at_infall", "jz_at_infall", "T_at_infall",
                      "ejection_a", "ejection_time",
                      "mass_at_ejection", "metals_at_ejection", "jz_at_ejection", "T_at_ejection",
                      "cycle_r_max_at", "cycle_r_max", "cycle_z_max_at", "cycle_z_max",
                      ]
    if isinstance(add_blocks, str):
        add_blocks = (add_blocks,)
    if units is None:
        # TODO: angmom units
        units = dict(TIME='a_form', MASS='Msol', TEMP='K',
                     ANGMOM=None, POS='kpc')
    gas = snap.root.gas
    environment.gc_full_collect()

    if isinstance(data, str):
        filename = data
        if environment.verbose >= environment.VERBOSE_TACITURN:
            print('reading the gas trace information from %s...' % filename)
        data = read_traced_gas(data)
    else:
        filename = '<given data>'

    # filter to gas only
    gas_type = {1, 2}
    data = dict([i for i in iter(data.items()) if i[1][0][0] in gas_type])

    if environment.verbose >= environment.VERBOSE_TALKY:
        print('test IDs and find matching IDs...')
    if len(set(data.keys()) - set(gas['ID'])) > 0:
        raise RuntimeError('Traced gas IDs in "%s" have ' % filename +
                           '%s ' % nice_big_num_str(len(
            set(data.keys()) - set(gas['ID']))) +
                           'elements that are not in the snapshot!')
    tracedIDs = (set(data.keys()) & set(gas['ID']))

    trmask = np.array([(ID in tracedIDs) for ID in gas['ID']], dtype=bool)
    if environment.verbose >= environment.VERBOSE_TALKY:
        print('  found %s (of %s)' % (nice_big_num_str(len(tracedIDs)),
                                      nice_big_num_str(len(data))), end=' ')
        print('traced IDs that are in the snapshot')

    def add_block(name, block):
        if name in add_blocks:
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('  "%s"' % name)
            gas[name] = block

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('adding the blocks:')

    trididx = np.argsort(list(data.keys()))
    gididx = np.argsort(gas['ID'])
    gididx_traced = gididx[trmask[gididx]]

    # type = not traced (0), in region (1), out of region (2)
    # number of full cycles (out and in)

    trace_type = UnitArr(np.zeros(len(gas), dtype=int))
    trtype = np.array([i[0][0] for i in data.values()])
    trace_type[gididx_traced] = trtype[trididx]
    add_block('trace_type', trace_type)
    del trace_type

    n_cyc = np.array([i[0][1] for i in data.values()])
    num_recycled = UnitArr(np.empty(len(gas), dtype=n_cyc.dtype))
    num_recycled[~trmask] = -1
    num_recycled[gididx_traced] = n_cyc[trididx]
    add_block('num_recycled', num_recycled)
    set_N_cycles = set(n_cyc)
    if environment.verbose >= environment.VERBOSE_TALKY:
        print('  +++ number of recycles that occured:', set_N_cycles, '+++')

    # Create blocks with shape (N, max(recycl)+1) for traced quantities at the
    # events of entering the region. N is the number of the gas particles. Each
    # particle has entries for each of their enter events, those events that do
    # not exists, get filled with `invalid`.

    max_N_cycle = max(set_N_cycles)
    infall_t = np.array([[e[0] for e in i[1:1 + 3 * n + 1:3]] + [invalid] * (max_N_cycle - n)
                         for n, i in zip(n_cyc, list(data.values()))])
    infall_a = UnitArr(np.empty((len(gas), max_N_cycle + 1),
                                dtype=infall_t.dtype), units=units['TIME'])
    infall_a[~trmask] = invalid
    infall_a[gididx_traced] = infall_t[trididx]
    add_block('infall_a', infall_a)
    del infall_t
    environment.gc_full_collect()

    from .snapshot import age_from_form
    # only convert reasonable values & ensure not to overwrite blocks
    mask = (infall_a != invalid) & np.isfinite(infall_a)
    infall_time = infall_a.copy()
    new = gas.cosmic_time() - age_from_form(infall_time[mask], subs=gas)
    infall_time.units = new.units
    infall_time[mask] = new
    del new
    add_block('infall_time', infall_time)
    del infall_time
    environment.gc_full_collect()

    for name, idx, unit in [('mass_at_infall', 1, units['MASS']),
                            ('metals_at_infall', 2, units['MASS']),
                            ('jz_at_infall', 3, units['ANGMOM']),
                            ('T_at_infall', 4, units['TEMP'])]:
        if name not in add_blocks:
            continue
        infall_Q = np.array([[e[idx] for e in i[1:1 + 3 * n + 1:3]] +
                             [invalid] * (max_N_cycle - n)
                             for n, i in zip(n_cyc, list(data.values()))])
        block = UnitArr(np.empty((len(gas), max_N_cycle + 1),
                                 dtype=infall_Q.dtype),
                        units=unit)
        block[~trmask] = invalid
        block[gididx_traced] = infall_Q[trididx]
        add_block(name, block)
        del infall_Q, block
        environment.gc_full_collect()

    # Create blocks with shape (N, max(recycl)+1) for traced quantities at the
    # events of ejection / leaving the region. N is the number of the gas
    # particles. Each particle has entries for each of their ejection events,
    # those events that do not exists, get filled with `invalid`.
    max_N_cycle = max(set_N_cycles)
    eject_t = np.array([[e[0] for e in i[2:2 + 3 * n:3]] +
                        [i[2 + 3 * n][0] if t == 2 else invalid] +
                        [invalid] * (max_N_cycle - n)
                        for n, t, i in zip(n_cyc, trtype, list(data.values()))])
    ejection_a = UnitArr(np.empty((len(gas), max_N_cycle + 1),
                                  dtype=eject_t.dtype),
                         units=units['TIME'])
    ejection_a[~trmask] = invalid
    ejection_a[gididx_traced] = eject_t[trididx]
    del eject_t
    add_block('ejection_a', ejection_a)
    environment.gc_full_collect()

    from .snapshot import age_from_form
    # only convert reasonable values & ensure not to overwrite blocks
    mask = (ejection_a != invalid) & np.isfinite(ejection_a)
    ejection_time = ejection_a.copy()
    new = gas.cosmic_time() - age_from_form(ejection_time[mask], subs=gas)
    ejection_time.units = new.units
    ejection_time[mask] = new
    del new
    add_block('ejection_time', ejection_time)
    environment.gc_full_collect()

    for name, idx, unit in [('mass_at_ejection', 1, units['MASS']),
                            ('metals_at_ejection', 2, units['MASS']),
                            ('jz_at_ejection', 3, units['ANGMOM']),
                            ('T_at_ejection', 4, units['TEMP'])]:
        if name not in add_blocks:
            continue
        eject_Q = np.array([[e[idx] for e in i[2:2 + 3 * n:3]] +
                            [i[2 + 3 * n][idx] if t == 2 else invalid] +
                            [invalid] * (max_N_cycle - n)
                            for n, t, i in zip(n_cyc, trtype, list(data.values()))])
        block = UnitArr(np.empty((len(gas), max_N_cycle + 1),
                                 dtype=eject_Q.dtype),
                        units=unit)
        block[~trmask] = invalid
        block[gididx_traced] = eject_Q[trididx]
        add_block(name, block)
        del eject_Q, block
        environment.gc_full_collect()

    # for each cycle ther is a maximum travel distance, plus one more for those
    # particles that are outside the region: store them
    for name, idx, unit in [('cycle_r_max_at', 0, units['TIME']),
                            ('cycle_r_max', 1, units['POS']),
                            ('cycle_z_max_at', 2, units['TIME']),
                            ('cycle_z_max', 3, units['POS'])]:
        if name not in add_blocks:
            continue
        pos = np.array([[e[idx] for e in i[3:3 + 3 * n:3]] +
                        [i[3 + 3 * n][idx] if t == 2 else invalid] +
                        [invalid] * (max_N_cycle - n)
                        for n, t, i in zip(n_cyc, trtype, list(data.values()))])
        block = UnitArr(np.empty((len(gas), max_N_cycle + 1),
                                 dtype=pos.dtype),
                        units=unit)
        block[~trmask] = invalid
        block[gididx_traced] = pos[trididx]
        add_block(name, block)
        del pos, block
        environment.gc_full_collect()

    if add_derived:
        fill_derived_gas_trace_qty(snap, units=units, invalid=invalid)


def fill_derived_gas_trace_qty(snap, units=None, invalid=0.0):
    """
    Derive blocks from the gas trace blocks and add them.

    Among the new blocks are:
      "out_time",
      "last_infall_a", "last_infall_time", "mass_at_last_infall",
        "metals_at_last_infall", "jz_at_last_infall", "T_at_last_infall",
      "last_ejection_a", "last_ejection_time", "mass_at_last_ejection",
        "metals_at_last_ejection", "jz_at_last_ejection", "T_at_last_ejection"'

    Note:
        The arguments `units` and `invalid` shall be the same as for
        `fill_gas_from_traced`.
        Also, some blocks added by `fill_gas_from_traced` are needed. Make sure
        they are avaiable!

    Args:
        snap (Snap):        The snapshot to fill with the data (has to be the one
                            at z=0 of the simulation used to create the trace
                            file).
        units (dict):       The units to use for masses, lengths, etc..
        invalid (float/NaN/inf):
                            The value to fill invalid entries with. Such entries
                            are properties of cycles that did not happen for a
                            given particle.
    """
    if units is None:
        # TODO: angmom units
        units = dict(TIME='a_form', MASS='Msol', TEMP='K',
                     ANGMOM=None, POS='kpc')
    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('adding blocks that can be derived from the gas trace blocks:')

    gas = snap.gas

    trmask = (gas['num_recycled'] != -1)
    set_N_cycles = set(gas['num_recycled'])
    max_N_cycle = max(set_N_cycles)

    # each (full) cycle takes some given time
    if max_N_cycle > 0:
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('  "out_time",')
        ejected = gas['ejection_time'][:, :-1]
        infall = gas['infall_time'][:, 1:]
        gas['out_time'] = infall - ejected
        mask = (ejected == invalid) | ~np.isfinite(ejected)
        del ejected
        environment.gc_full_collect()
        gas['out_time'][mask] = invalid
        mask = (infall == invalid) | ~np.isfinite(infall)
        del infall
        environment.gc_full_collect()
        gas['out_time'][mask] = invalid
    environment.gc_full_collect()

    # The events of the last infall and the last ejection are a bit messy to
    # access. Create extra blocks:
    # last_infall_idx = np.sum(~np.isnan(gas['infall_a']), axis=-1) - 1
    last_infall_idx = np.sum(gas['infall_a'] != invalid, axis=-1) - 1
    last_infall_idx = np.arange(len(last_infall_idx)), last_infall_idx
    for last, alle in [('last_infall_a', 'infall_a'),
                       ('last_infall_time', 'infall_time'),
                       ('mass_at_last_infall', 'mass_at_infall'),
                       ('metals_at_last_infall', 'metals_at_infall'),
                       ('jz_at_last_infall', 'jz_at_infall'),
                       ('T_at_last_infall', 'T_at_infall')]:
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('  "%s"' % last)
        gas[last] = UnitArr(np.empty(len(gas), dtype=gas[alle].dtype),
                            units=gas[alle].units)
        gas[last][~trmask] = invalid
        gas[last][trmask] = gas[alle][last_infall_idx][trmask]
    del last_infall_idx
    environment.gc_full_collect()

    # last_ejection_idx = np.sum(~np.isnan(gas['ejection_a']), axis=-1) - 1
    last_ejection_idx = np.sum(gas['ejection_a'] != invalid, axis=-1) - 1
    last_ejection_idx = np.arange(len(last_ejection_idx)), last_ejection_idx
    for last, alle in [('last_ejection_a', 'ejection_a'),
                       ('last_ejection_time', 'ejection_time'),
                       ('mass_at_last_ejection', 'mass_at_ejection'),
                       ('metals_at_last_ejection', 'metals_at_ejection'),
                       ('jz_at_last_ejection', 'jz_at_ejection'),
                       ('T_at_last_ejection', 'T_at_ejection')]:
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('  "%s"' % last)
        gas[last] = UnitArr(np.empty(len(gas), dtype=gas[alle].dtype),
                            units=gas[alle].units)
        gas[last][~trmask] = invalid
        gas[last][trmask] = gas[alle][last_ejection_idx][trmask]
    del last_ejection_idx
    environment.gc_full_collect()

    """
    inmask  = (gas['trace_type'] == 1)
    outmask = (gas['trace_type'] == 2)

    gididx_in = gididx[inmask[gididx]]
    gididx_out = gididx[outmask[gididx]]
    trididx_in = trididx[(trtype==1)[trididx]]
    trididx_out = trididx[(trtype==2)[trididx]]

    # Amount of metals gained nside and outside the galaxy.
    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('  "metal_gain_in", "metal_gain_out"')
    # all traced particles have a value for metals gained before entering:
    gas['metal_gain_in'] = UnitArr( np.zeros(len(gas), dtype=float),
                                    units=units['MASS'] )
    gas['metal_gain_out'] = gas['metals_at_infall'][:,0].in_units_of(units['MASS'],copy=True)
    gas['metal_gain_in'][~trmask] = np.nan
    gas['metal_gain_out'][~trmask] = np.nan
    gas['metal_gain_in'][~trmask] = invalid
    gas['metal_gain_out'][~trmask] = invalid
    metals_last_enter = np.empty( len(data), dtype=float )
    metals_last_enter[trididx] = \
            gas['metals_at_infall'][:,0][gididx_traced].in_units_of(units['MASS'],copy=True)
    # some had gained more in (full) re-cycles (with no recycles: done!)
    for n in sorted(set_N_cycles)[1:]:  # 1,2,3,4,...
        gididx_n = gididx[(gas['num_recycled']==n)[gididx]]
        trididx_n = trididx[(n_cyc==n)[trididx]]
        # leaving the region
        idx = 2 + (n-1)*3
        metals_leave = np.array( [(i[idx][2] if i[0][1]>=n else invalid)
                                  for i in data.itervalues()] )
        idx = 1 + n*3
        metals_enter = np.array( [(i[idx][2] if i[0][1]>=n else invalid)
                                  for i in data.itervalues()] )
        gas['metal_gain_in'][gididx_n] += (metals_leave-metals_last_enter)[trididx_n]
        gas['metal_gain_out'][gididx_n] += (metals_enter-metals_leave)[trididx_n]
        metals_last_enter = metals_enter
    del gididx_n, trididx_n, metals_enter, metals_leave, metals_last_enter
    environment.gc_full_collect()
    gasmetals = gas['metals'].in_units_of(units['MASS'],subs=snap)
    # ... those that are currently outside, have gained something since they left
    # the region for the last time
    metals_last_leave = np.array( [(i[2+i[0][1]*3][2] if i[0][0]==2 else invalid)
                                   for i in data.itervalues()] )
    gas['metal_gain_out'][gididx_out] += gasmetals[gididx_out] - metals_last_leave[trididx_out]
    # ... and have gained something during they have been inside the region (which
    # was not a full cycle!)
    metals_last_enter = np.array( [i[1+i[0][1]*3][2] for i in data.itervalues()] )
    gas['metal_gain_in'][gididx_out] += \
            metals_last_leave[trididx_out] - metals_last_enter[trididx_out]
    # ... and those that are currently inside, have gained something since the
    # last time they entered
    gas['metal_gain_in'][gididx_in] += gasmetals[gididx_in] - metals_last_enter[trididx_in]
    del gasmetals, metals_last_enter, metals_last_leave
    environment.gc_full_collect()
    gas['metal_gain_out'].convert_to(gas['metals'].units)
    gas['metal_gain_in'].convert_to(gas['metals'].units)
    """


def read_pyigm_COS_halos_EWs(transitions=('HI 1215', 'MgII 2796', 'MgII 2803', 'CIV 1548',
                                          'CIV 1550', 'OVI 1031', 'OVI 1037')):
    '''
    Read the EW as a function of impact parameter of the COS halos using pyigm.

    References: Tumlinson+11; Werk+12; Tumlinson+13; Werk+13; Werk+14

    Args:
        transitions (iterable): The line transition to load the data for.
                                Eg. 'HI 1215' for Lyman-alpha. Or 'HI 1025',
                                'MgII 2796', 'MgII 2803', 'OVI 1037', 'OVI 1031',
                                'SiIII 1206', 'CIII 977', 'CIV 1548', 'CIV 1550',
                                etc.

    Returns:
        data (dict):            Relevant data for each line of sight. It includes:
                                'z' (reshifts), 'rho' (impact parameter),
                                'EW' (equivalent widths of the lines),
                                'Mstars' (stellar masses of the associated
                                galaxies), and more. Each entry is a UnitArr.
    '''
    import pyigm
    from pyigm.cgm import cos_halos
    cos_halos = cos_halos.COSHalos()
    data = {trans: [] for trans in transitions}
    for sys in cos_halos:
        sys = sys.to_dict()
        Mhalo = sys['galaxy']['halo_mass']
        Mstars = sys['galaxy']['stellar_mass']
        sSFR = sys['galaxy']['ssfr']
        Rvir = sys['galaxy']['rvir']
        for component in sys['igm_sys']['components']:
            for line in sys['igm_sys']['components'][component]['lines']:
                tname = sys['igm_sys']['components'][component]['lines'][line]['name']
                if tname in transitions:
                    EW = sys['igm_sys']['components'][component]['lines'][line]['attrib']['EW']['value']
                    data[tname].append([sys['z'], sys['rho'], EW,
                                        Mhalo, Mstars, sSFR, Rvir])
    for trans, d in data.items():
        if len(d) == 0:
            raise RuntimeError('transition "%s" not found' % trans)

    data = {trans: np.array(d) for trans, d in data.items()}
    return {trans: {
        'z': UnitArr(d[:, 0]),
        'rho': UnitArr(d[:, 1], 'kpc'),
        # 'EW':       UnitArr(10.**d[:,2], 'Angstrom'),
        'EW': UnitArr(d[:, 2], 'Angstrom'),
        'Mhalo': UnitArr(10. ** d[:, 3], 'Msol'),
        'Mstars': UnitArr(10. ** d[:, 4], 'Msol'),
        'sSFR': UnitArr(d[:, 5]),
        'Rvir': UnitArr(d[:, 6], 'kpc'),
    } for trans, d in data.items()}




"""
Binning into some 1D grid (e.g. profiles).

Doctests:
    >>> from ..transformation import *
    >>> from ..snapshot import *
    >>> from .mapping import *
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_470', physical=True)
    >>> Translation(UnitArr([-48087.1,-49337.1,-46084.3],'kpc')).apply(s)
    >>> s['vel'] -= UnitArr([-42.75,-15.60,-112.20],'km s**-1')
    load block vel... done.
    >>> ext = UnitArr( [[-200,200],[-200,200]], 'kpc' )
    >>> m = map_qty(s.gas, ext, True, 'rho', Npx=128) # doctest: +ELLIPSIS
    load block rho... done.
    ...
    done with SPH grid
    >>> prof, r_edges = profile_from_map(m, ext, reduction='mean', Nbins=40)
    >>> np.log10(prof) # doctest: +ELLIPSIS
    UnitArr([6.52927349, 6.69610836, 6.72196057, 6.70476286, 6.6541627 ,
             6.64614539, 6.42186697, 5.93403278, 5.85748263, 5.88598258,
             5.87697597, 5.81081813, 5.77023788, 5.75011381, 5.71452994,
             5.68998774, 5.66865784, 5.6529112 , 5.63884698, 5.62835883,
             5.61282864, 5.59369405, 5.57579912, 5.56786731, 5.55873967,
             5.54992073, 5.53984417, 5.52930242, 5.51920267, 5.51656575,
             5.51265098, 5.50450026, 5.4986523 , 5.48687964, 5.47931559,
             5.46554878, 5.45591616, 5.44801523, 5.44204756, 5.43046586])
"""
__all__ = ['profile_from_map']

import numpy as np
from ..units import UnitArr, UnitQty
from ..utils import dist, weighted_percentile

def profile_from_map(m, extent, av=None, Nbins=None, reduction='sum',
                     ref_center=None, surface_dens=False):
    """
    Create a profile from a 2D grid.

    Args:
        m (array-like):             The map to create the profile from.
        extent (array-like):        The extent of the map.
        av (array-like):            Another map used to weigh the maps values.
        Nbins (int):                The number of radial bins. Defaults to
                                    `min(*m.shape) / 3 + 1`.
        reduction (str):            The method to combine the pixel values into
                                    the bins. Available are:
                                    * 'sum':    Simply sum up the pixel values.
                                    * 'mean':   Take the mean/average of the
                                                pixels.
                                    * 'median': Take the median of the pixels.
                                    * 'std':    Calculate the standard deviation
                                                if the pixels in the radial bin.
        ref_center (array-like):    The reference center, i.e. the point from
                                    which to calculate the radii. Defaults to the
                                    center of the map.
        TODO

    Returns:
        profile (UnitQty):  The profile values for each bin.
        r_edges (UnitQty):  The bin edges.
    """
    # prepare arguments
    m = UnitQty(m)
    if len(m.shape) != 2:
        raise ValueError("The map is not 2D!")
    if av is not None:
        av = UnitQty(av)
        if av.shape != m.shape:
            raise ValueError("The 'average map' has different shape than the "
                             "main map!")
    if Nbins is None:
        Nbins = min(*m.shape) / 3 + 1
    Nbins = int(Nbins)
    extent = UnitQty(extent)
    if extent.shape != (2,2):
        raise ValueError("`extent` has to have shape (2,2)!")
    if ref_center is None:
        ref_center = np.sum(extent,axis=1) / 2.
    ref_center = UnitQty(ref_center, units=extent.units)
    if ref_center.shape != (2,):
        raise ValueError("`ref_center` has to have shape (2,)!")

    # pixel center positions and radii:
    X = np.linspace( float(extent[0,0]), float(extent[0,1]), m.shape[0]+1 )
    Y = np.linspace( float(extent[1,0]), float(extent[1,1]), m.shape[1]+1 )
    X = (X[:-1] + X[1:]) / 2.
    Y = (Y[:-1] + Y[1:]) / 2.
    X,Y = np.meshgrid(X,Y)
    pos = np.swapaxes( np.dstack((X,Y)), 0, 1 )
    del X, Y
    r_px = dist(pos.reshape((-1,2)), ref_center).reshape(pos.shape[:2])

    # radial bins
    #r_edges = np.linspace( float(r_px.min()), float(r_px.max()), Nbins+1 )
    r_edges = np.linspace( 0, float(np.min(extent[:,1]-extent[:,0])/2.), Nbins+1 )
    r_edges = UnitArr(r_edges, extent.units)
    #dr = UnitScalar(r_edges[1] - r_edges[0], r_edges.units)
    idx = np.digitize(r_px.flatten(), r_edges).reshape(r_px.shape)

    # reduce map into bins:
    if av is None:
        if reduction in ['sum', 'mean', 'median', 'std']:
            red_func = getattr(np, reduction)
        else:
            raise ValueError("Unknown reduction method '%s'!" % reduction)
        profile = [ red_func(m[idx==i]) for i in range(1,Nbins+1) ]
    else:
        if reduction == 'mean':
            red_func = lambda q,a: np.average(q,weights=a)
        elif reduction == 'median':
            red_func = lambda q,a: weighted_percentile(q,[50],weights=a)[0]
        elif reduction in ['sum', 'mean', 'median', 'std']:
            raise ValueError("Cannot use weights with reduction method "
                             "'%s'!" % reduction)
        else:
            raise ValueError("Unknown reduction method '%s'!" % reduction)
        profile = [ red_func(m[idx==i],av[idx==i]) for i in range(1,Nbins+1) ]
    profile = UnitArr(profile, m.units)

    if surface_dens:
        ring_area = np.pi * (r_edges[1:]**2 - r_edges[:-1]**2)
        profile /= ring_area

    return profile, r_edges

